"""
This plugin searches for Dynatrace API Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class DynatraceApiTokenDetector(RegexBasedDetector):
    """Scans for Dynatrace API Tokens."""

    secret_type = "Dynatrace API Token"

    denylist = [
        # Dynatrace API Token
        re.compile(r"""(?i)dt0c01\.[a-z0-9]{24}\.[a-z0-9]{64}"""),
    ]



"""
This plugin searches for Squarespace Access Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class SquarespaceAccessTokenDetector(RegexBasedDetector):
    """Scans for Squarespace Access Tokens."""

    secret_type = "Squarespace Access Token"

    denylist = [
        re.compile(
            r"""(?i)(?:squarespace)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        )
    ]



"""
This plugin searches for Finnhub Access Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class FinnhubAccessTokenDetector(RegexBasedDetector):
    """Scans for Finnhub Access Tokens."""

    secret_type = "Finnhub Access Token"

    denylist = [
        # Finnhub Access Token
        re.compile(
            r"""(?i)(?:finnhub)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9]{20})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



from typing import Dict, Optional, Sequence

from llm_guard.exception import LLMGuardValidationError
from llm_guard.transformers_helpers import pipeline
from llm_guard.util import get_logger

from .base import Scanner

LOGGER = get_logger(__name__)

# This model was trained on a mixture of 33 datasets and 389 classes reformatted in the universal NLI format.
# The model is English only. You can also use it for multilingual zeroshot classification by first machine translating texts to English.
MODEL_LARGE = {
    "path": "MoritzLaurer/deberta-v3-large-zeroshot-v1.1-all-33",
    "onnx_path": "MoritzLaurer/deberta-v3-large-zeroshot-v1.1-all-33",
    "max_length": 512,
}
# This is essentially the same as its larger sister MoritzLaurer/deberta-v3-large-zeroshot-v1.1-all-33 only that it's smaller.
# Use it if you need more speed. The model is English-only.
MODEL_BASE = {
    "path": "MoritzLaurer/deberta-v3-base-zeroshot-v1.1-all-33",
    "onnx_path": "MoritzLaurer/deberta-v3-base-zeroshot-v1.1-all-33",
    "max_length": 512,
}
# Same as above, just smaller/faster.
MODEL_XSMALL = {
    "path": "MoritzLaurer/deberta-v3-xsmall-zeroshot-v1.1-all-33",
    "onnx_path": "MoritzLaurer/deberta-v3-xsmall-zeroshot-v1.1-all-33",
    "max_length": 512,
}
# Same as above, just even faster. The model only has 22 million backbone parameters.
# The model is 25 MB small (or 13 MB with ONNX quantization).
MODEL_XTREMEDISTIL = {
    "path": "MoritzLaurer/xtremedistil-l6-h256-zeroshot-v1.1-all-33",
    "onnx_path": "MoritzLaurer/xtremedistil-l6-h256-zeroshot-v1.1-all-33",
    "max_length": 512,
}

ALL_MODELS = [MODEL_LARGE, MODEL_BASE, MODEL_XSMALL, MODEL_XTREMEDISTIL]


class BanTopics(Scanner):
    """
    BanTopics class is used to ban certain topics from the prompt.

    It uses a HuggingFace model to perform zero-shot classification.
    """

    def __init__(
        self,
        topics: Sequence[str],
        *,
        threshold: float = 0.6,
        model: Optional[Dict] = None,
        use_onnx: bool = False,
        transformers_kwargs: Optional[Dict] = None,
    ):
        """
        Initialize BanTopics object.

        Parameters:
            topics (Sequence[str]): List of topics to ban.
            threshold (float, optional): Threshold to determine if a topic is present in the prompt. Default is 0.75.
            model (Dict, optional): Model to use for zero-shot classification. Default is deberta-v3-base-zeroshot-v1.
            use_onnx (bool, optional): Whether to use ONNX for inference. Default is False.
            transformers_kwargs (Dict, optional): Additional kwargs to pass to the transformers pipeline. Default is None.

        Raises:
            ValueError: If no topics are provided.
        """
        if model is None:
            model = MODEL_BASE

        if model not in ALL_MODELS:
            raise LLMGuardValidationError(f"Model must be in the list of allowed: {ALL_MODELS}")

        self._topics = topics
        self._threshold = threshold

        transformers_kwargs = transformers_kwargs or {}
        transformers_kwargs["max_length"] = model["max_length"]
        transformers_kwargs["truncation"] = True

        self._classifier = pipeline(
            task="zero-shot-classification",
            model=model["path"],
            onnx_model=model["onnx_path"],
            use_onnx=use_onnx,
            **transformers_kwargs,
        )

    def scan(self, prompt: str) -> (str, bool, float):
        if prompt.strip() == "":
            return prompt, True, 0.0

        output_model = self._classifier(prompt, self._topics, multi_label=False)

        max_score = round(max(output_model["scores"]) if output_model["scores"] else 0, 2)
        if max_score > self._threshold:
            LOGGER.warning(
                "Topics detected for the prompt",
                labels=output_model["labels"],
                scores=output_model["scores"],
            )

            return prompt, False, max_score

        LOGGER.debug(
            "No banned topics detected",
            labels=output_model["labels"],
            scores=output_model["scores"],
        )

        return prompt, True, 0.0



"""
This plugin searches for Prefect API Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class PrefectApiTokenDetector(RegexBasedDetector):
    """Scans for Prefect API Tokens."""

    secret_type = "Prefect API Token"

    denylist = [re.compile(r"""(?i)\b(pnu_[a-z0-9]{36})(?:['|\"|\n|\r|\s|\x60|;]|$)""")]



import re
from typing import Dict, List, Optional, Sequence

from llm_guard.exception import LLMGuardValidationError
from llm_guard.transformers_helpers import pipeline
from llm_guard.util import calculate_risk_score, get_logger

from .base import Scanner

LOGGER = get_logger(__name__)

_model_path = "philomath-1209/programming-language-identification"

SUPPORTED_LANGUAGES = [
    "ARM Assembly",
    "AppleScript",
    "C",
    "C#",
    "C++",
    "COBOL",
    "Erlang",
    "Fortran",
    "Go",
    "Java",
    "JavaScript",
    "Kotlin",
    "Lua",
    "Mathematica/Wolfram Language",
    "PHP",
    "Pascal",
    "Perl",
    "PowerShell",
    "Python",
    "R",
    "Ruby",
    "Rust",
    "Scala",
    "Swift",
    "Visual Basic .NET",
    "jq",
]


class Code(Scanner):
    """
    A class for scanning if the prompt includes code in specific programming languages.

    This class uses the transformers library to detect code snippets in the output of the language model.
    It can be configured to allow or block specific programming languages.
    """

    def __init__(
        self,
        languages: Sequence[str],
        *,
        is_blocked: bool = True,
        threshold: float = 0.5,
        use_onnx: bool = False,
        transformers_kwargs: Optional[Dict] = None,
    ):
        """
        Initializes Code with the allowed and denied languages.

        Parameters:
            languages (Sequence[str]): The list of programming languages to allow or deny.
            is_blocked (bool): Whether the languages are blocked or allowed. Default is True.
            threshold (float): The threshold for the risk score. Default is 0.5.
            use_onnx (bool): Whether to use ONNX for inference. Default is False.
            transformers_kwargs (Optional[Dict]): Optional keyword arguments for the transformers pipeline.

        Raises:
            LLMGuardValidationError: If the languages are not a subset of SUPPORTED_LANGUAGES.
        """
        if not set(languages).issubset(set(SUPPORTED_LANGUAGES)):
            raise LLMGuardValidationError(f"Languages must be a subset of {SUPPORTED_LANGUAGES}")

        self._languages = languages
        self._is_blocked = is_blocked
        self._threshold = threshold

        transformers_kwargs = transformers_kwargs or {}
        transformers_kwargs["truncation"] = True

        self._pipeline = pipeline(
            task="text-classification",
            model=_model_path,
            onnx_model=_model_path,
            use_onnx=use_onnx,
            **transformers_kwargs,
        )

        self._fenced_code_regex = re.compile(r"```(?:[a-zA-Z0-9]*\n)?(.*?)```", re.DOTALL)
        self._inline_code_regex = re.compile(r"`(.*?)`")

    def _extract_code_blocks(self, markdown: str) -> List[str]:
        # Extract fenced code blocks (between triple backticks)
        fenced_code_blocks = [
            block.strip() for block in self._fenced_code_regex.findall(markdown) if block.strip()
        ]

        # Extract inline code (between single backticks)
        inline_code = [
            code.strip()
            for code in self._inline_code_regex.findall(markdown)
            if code.strip() and any(char in code for char in "{}[]()=+-*/<>!")
        ]

        return fenced_code_blocks + inline_code

    def scan(self, prompt: str) -> (str, bool, float):
        if prompt.strip() == "":
            return prompt, True, 0.0

        # Try to extract code snippets from Markdown
        code_blocks = self._extract_code_blocks(prompt)
        if len(code_blocks) == 0:
            LOGGER.debug("No Markdown code blocks found in the output")
            return prompt, True, 0.0

        LOGGER.debug("Code blocks found in the output", code_blocks=code_blocks)

        # Only check when the code is detected
        for code_block in code_blocks:
            languages = self._pipeline(code_block)
            LOGGER.debug(
                "Detected languages in the code", languages=languages, code_block=code_block
            )

            for language in languages:
                language_name = language["label"]
                score = round(language["score"], 2)

                if score < self._threshold or language_name not in self._languages:
                    continue

                if self._is_blocked:
                    LOGGER.warning(
                        "Language is not allowed", language_name=language_name, score=score
                    )
                    return prompt, False, calculate_risk_score(score, self._threshold)

                if not self._is_blocked:
                    LOGGER.debug("Language is allowed", language_name=language_name, score=score)
                    return prompt, True, 0.0

        if self._is_blocked:
            LOGGER.debug("No blocked languages detected")
            return prompt, True, 0.0

        LOGGER.warning("No allowed languages detected")
        return prompt, False, 1.0



"""
This plugin searches for Bitbucket Client ID and Client Secret
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class BitbucketDetector(RegexBasedDetector):
    """Scans for Bitbucket Client ID and Client Secret."""

    secret_type = "Bitbucket Secrets"

    denylist = [
        # For Bitbucket Client ID
        re.compile(
            r"""(?i)(?:bitbucket)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9]{32})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
        # For Bitbucket Client Secret
        re.compile(
            r"""(?i)(?:bitbucket)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9=_\-]{64})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



import importlib
from functools import lru_cache
from typing import Literal, Optional, get_args

from .exception import LLMGuardValidationError
from .util import device, get_logger, lazy_load_dep

LOGGER = get_logger(__name__)


@lru_cache(maxsize=None)  # Set maxsize=None for an unbounded cache
def get_tokenizer(model_identifier: str):
    """
    This function loads a tokenizer given a model identifier and caches it.
    Subsequent calls with the same model_identifier will return the cached tokenizer.

    Args:
        model_identifier (str): The model identifier to load the tokenizer for.
    """
    transformers = lazy_load_dep("transformers")
    tokenizer = transformers.AutoTokenizer.from_pretrained(model_identifier)
    return tokenizer


@lru_cache(maxsize=None)  # Unbounded cache
def is_onnx_supported() -> bool:
    is_supported = importlib.util.find_spec("optimum.onnxruntime") is not None
    if not is_supported:
        LOGGER.warning(
            "ONNX Runtime is not available. "
            "Please install optimum: "
            "`pip install llm-guard[onnxruntime]` for CPU or "
            "`pip install llm-guard[onnxruntime-gpu]` for GPU to enable ONNX Runtime optimizations."
        )

    return is_supported


def _ort_model_for_sequence_classification(model: str, export: bool = False, subfolder: str = ""):
    if device().type == "cuda":
        optimum_onnxruntime = lazy_load_dep("optimum.onnxruntime", "optimum[onnxruntime-gpu]")
        tf_model = optimum_onnxruntime.ORTModelForSequenceClassification.from_pretrained(
            model,
            export=export,
            subfolder=subfolder,
            file_name="model.onnx",
            provider="CUDAExecutionProvider",
            use_io_binding=True,
        )

        LOGGER.debug("Initialized classification ONNX model", model=model, device=device())

        return tf_model

    optimum_onnxruntime = lazy_load_dep("optimum.onnxruntime", "optimum[onnxruntime]")
    tf_model = optimum_onnxruntime.ORTModelForSequenceClassification.from_pretrained(
        model,
        export=export,
        subfolder=subfolder,
        file_name="model.onnx",
    )
    LOGGER.debug("Initialized classification ONNX model", model=model, device=device())

    return tf_model


def get_tokenizer_and_model_for_classification(
    model: str, onnx_model: Optional[str] = None, use_onnx: bool = False, **kwargs
):
    """
    This function loads a tokenizer and model given a model identifier and caches them.
    Subsequent calls with the same model_identifier will return the cached tokenizer.

    Args:
        model (str): The model identifier to load the tokenizer and model for.
        onnx_model (Optional[str]): The model identifier to load the ONNX model for. Defaults to None.
        use_onnx (bool): Whether to use the ONNX version of the model. Defaults to False.
        **kwargs: Keyword arguments to pass to the tokenizer and model.
    """
    tf_tokenizer = get_tokenizer(model)
    transformers = lazy_load_dep("transformers")

    if kwargs.get("max_length", None) is None:
        kwargs["max_length"] = tf_tokenizer.model_max_length

    if use_onnx and is_onnx_supported() is False:
        LOGGER.warning("ONNX is not supported on this machine. Using PyTorch instead of ONNX.")
        use_onnx = False

    if use_onnx is False:
        tf_model = transformers.AutoModelForSequenceClassification.from_pretrained(model)
        LOGGER.debug("Initialized classification model", model=model, device=device())

        return tf_tokenizer, tf_model

    subfolder = "onnx" if onnx_model == model else ""
    if onnx_model is not None:
        model = onnx_model

    # Hack for some models
    tf_tokenizer.model_input_names = ["input_ids", "attention_mask"]

    tf_model = _ort_model_for_sequence_classification(
        model, export=onnx_model is None, subfolder=subfolder
    )

    return tf_tokenizer, tf_model


ClassificationTask = Literal["text-classification", "zero-shot-classification", "ner"]


def pipeline(
    task: str, model: str, onnx_model: Optional[str] = None, use_onnx: bool = False, **kwargs
):
    if task not in get_args(ClassificationTask):
        raise LLMGuardValidationError(f"Invalid task. Must be one of {ClassificationTask}")

    if task == "ner":
        return _pipeline_ner(model, onnx_model, use_onnx, **kwargs)

    transformers = lazy_load_dep("transformers")
    tf_tokenizer, tf_model = get_tokenizer_and_model_for_classification(
        model, onnx_model, use_onnx, **kwargs
    )

    return transformers.pipeline(
        task,
        model=tf_model,
        tokenizer=tf_tokenizer,
        device=device(),
        batch_size=1,
        **kwargs,
    )


def _pipeline_ner(model: str, onnx_model: Optional[str] = None, use_onnx: bool = False, **kwargs):
    transformers = lazy_load_dep("transformers")
    tf_tokenizer = get_tokenizer(model)

    if use_onnx and is_onnx_supported() is False:
        LOGGER.warning("ONNX is not supported on this machine. Using PyTorch instead of ONNX.")
        use_onnx = False

    if use_onnx:
        subfolder = "onnx" if onnx_model == model else ""
        if onnx_model is not None:
            model = onnx_model

        optimum_onnxruntime = lazy_load_dep(
            "optimum.onnxruntime",
            "optimum[onnxruntime]" if device().type != "cuda" else "optimum[onnxruntime-gpu]",
        )
        tf_model = optimum_onnxruntime.ORTModelForTokenClassification.from_pretrained(
            model,
            export=onnx_model is None,
            subfolder=subfolder,
            provider="CUDAExecutionProvider" if device().type == "cuda" else "CPUExecutionProvider",
            use_io_binding=True if device().type == "cuda" else False,
        )
        LOGGER.debug("Initialized NER ONNX model", model=model, device=device())
    else:
        tf_model = transformers.AutoModelForTokenClassification.from_pretrained(model)
        LOGGER.debug("Initialized NER model", model=model, device=device())

    return transformers.pipeline(
        "ner",
        model=tf_model,
        tokenizer=tf_tokenizer,
        device=device(),
        batch_size=1,
        **kwargs,
    )



import importlib
import json
import logging
import re
import sys
from functools import lru_cache
from typing import Any, Dict, List, Literal, Optional

import structlog

LOGGER = structlog.getLogger(__name__)

"""
This file contains utility functions.
This is meant for internal use and not part of the public API.
"""

LOG_LEVELS = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
EXTERNAL_LOGGERS = {
    "transformers",
}


def configure_logger(log_level: LOG_LEVELS = "INFO"):
    """
    Configures the logger for the package.

    Args:
        log_level: The log level to use for the logger. It should be one of the following strings:
            "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL".
    """
    logging.basicConfig(
        format="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
        level=log_level,
        stream=sys.stdout,
    )
    structlog.configure(logger_factory=structlog.stdlib.LoggerFactory())
    for log_name in EXTERNAL_LOGGERS:
        logging.getLogger(log_name).setLevel(logging.WARNING)


def _get_library_name() -> str:
    return __name__.split(".")[0]


def _get_library_root_logger() -> logging.Logger:
    return logging.getLogger(_get_library_name())


def get_logger(name: Optional[str] = None) -> Any:
    """
    Return a logger with the specified name.
    """

    if name is None:
        name = _get_library_name()

    return structlog.getLogger(name)


# Detect pytorch device
@lru_cache(maxsize=None)  # Unbounded cache
def device():
    torch = lazy_load_dep("torch")
    if torch.cuda.is_available():
        return torch.device("cuda:0")
    elif torch.backends.mps.is_available():
        return torch.device("mps")

    return torch.device("cpu")


def read_json_file(json_path: str) -> Dict[str, List[str]]:
    """
    Reads a JSON file and returns its contents as a Python dictionary.

    Args:
        json_path: The path to the JSON file to be read.

    Returns:
        A dictionary representation of the JSON file's contents. If an error occurs (e.g. file not found or JSON decoding error),
        an empty dictionary is returned and an error message is logged.

    Raises:
        FileNotFoundError: If the provided json_path does not point to an existing file.
        json.decoder.JSONDecodeError: If the provided file cannot be parsed as JSON.
    """

    result = {}
    try:
        with open(json_path, "r") as myfile:
            result = json.load(myfile)
            LOGGER.debug("Loaded json file", path=json_path)
    except FileNotFoundError:
        LOGGER.error("Could not find file", path=json_path)
    except json.decoder.JSONDecodeError as json_error:
        LOGGER.error("Could not parse file", path=json_path, error=json_error)
    return result


def combine_json_results(results: Dict[str, List[str]]) -> List[str]:
    """
    Combines values from a dictionary with list values into a single list.

    Args:
       results: A dictionary where values are lists.

    Returns:
       A list containing all the values from the input dictionary.
    """

    all_items = []
    for item in results:
        all_items.extend(results[item])
    return all_items


def lazy_load_dep(import_name: str, package_name: Optional[str] = None):
    """Helper function to lazily load optional dependencies. If the dependency is not
    present, the function will raise an error _when used_.

    NOTE: This wrapper adds a warning message at import time.
    """

    if package_name is None:
        package_name = import_name

    spec = importlib.util.find_spec(import_name)
    if spec is None:
        LOGGER.warning(
            f"Optional feature dependent on missing package: {import_name} was initialized.\n"
            f"Use `pip install {package_name}` to install the package if running locally."
        )

    return importlib.import_module(import_name)


def calculate_risk_score(score: float, threshold: float) -> float:
    if score > threshold:
        return 1.0

    risk_score = round(abs(score - threshold) / threshold, 1)
    # Ensure risk score is between 0 and 1
    return min(max(risk_score, 0), 1)


def chunk_text(text: str, chunk_size: int) -> List[str]:
    text = text.strip()
    return [text[i : i + chunk_size] for i in range(0, len(text), chunk_size)]


def chunk_text_by_sentences(text: str, max_chunk_size: int) -> List[str]:
    nltk = lazy_load_dep("nltk")

    try:
        nltk.data.find("tokenizers/punkt")
    except LookupError:
        nltk.download("punkt")

    sentences = nltk.sent_tokenize(text.strip())

    chunks = []
    chunk = []
    chunk_size = 0

    for sentence in sentences:
        sentence_length = len(sentence)
        if chunk_size + sentence_length <= max_chunk_size:
            chunk.append(sentence)
            chunk_size += sentence_length
        else:
            if chunk:  # Check if chunk is non-empty
                chunks.append(" ".join(chunk))
            chunk = [sentence]
            chunk_size = sentence_length

    if chunk:  # Don't forget the last chunk, and check if it's non-empty
        chunks.append(" ".join(chunk))

    return [chunk for chunk in chunks if chunk.strip()]


def split_text_by_sentences(text: str) -> List[str]:
    nltk = lazy_load_dep("nltk")

    try:
        nltk.data.find("tokenizers/punkt")
    except LookupError:
        nltk.download("punkt")

    return nltk.sent_tokenize(text.strip())


url_pattern = re.compile(
    r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+"
)


def extract_urls(text: str) -> List[str]:
    """
    Extracts URLs from the given text.
    """
    return url_pattern.findall(text)



"""
This plugin searches for MessageBird API tokens and client IDs.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class MessageBirdDetector(RegexBasedDetector):
    """Scans for MessageBird secrets."""

    secret_type = "MessageBird Secret"

    denylist = [
        # MessageBird API Token
        re.compile(
            r"""(?i)(?:messagebird|message-bird|message_bird)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9]{25})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
        # MessageBird Client ID
        re.compile(
            r"""(?i)(?:messagebird|message-bird|message_bird)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



from typing import Sequence, Union

from llm_guard.input_scanners.regex import MatchType
from llm_guard.input_scanners.regex import Regex as InputRegex

from .base import Scanner


class Regex(Scanner):
    """
    A class used to detect patterns in the output of a language model using regular expressions.

    This class relies on the list of regular expressions provided by the user. If any of the patterns
    matches the output, the output is considered invalid. It is also possible to redact the output.
    """

    def __init__(
        self,
        patterns: Sequence[str],
        *,
        is_blocked: bool = True,
        match_type: Union[MatchType, str] = MatchType.SEARCH,
        redact=True,
    ):
        """
        Initializes an instance of the Regex class.

        Parameters:
            patterns (Sequence[str]): A list of regular expressions to use for pattern matching.
            is_blocked (bool): Whether the patterns are blocked or allowed.
            match_type (str): The type of match to use. Can be either "search" or "fullmatch".
            redact (bool): Whether to redact the output or not.

        Raises:
            ValueError: If no patterns provided or both good and bad patterns provided.
        """

        self._scanner = InputRegex(
            patterns, is_blocked=is_blocked, match_type=match_type, redact=redact
        )

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        return self._scanner.scan(output)



from llm_guard.input_scanners.sentiment import Sentiment as InputSentiment
from llm_guard.input_scanners.sentiment import _lexicon

from .base import Scanner


class Sentiment(Scanner):
    """
    A sentiment scanner based on the NLTK's SentimentIntensityAnalyzer. It is used to detect if a model output
    has a sentiment score lower than the threshold, indicating a negative sentiment.
    """

    def __init__(self, *, threshold: float = -0.1, lexicon: str = _lexicon):
        """
        Initializes Sentiment with a threshold and a chosen lexicon.

        Parameters:
           threshold (float): Threshold for the sentiment score (from -1 to 1). Default is -0.1.
           lexicon (str): Lexicon for the SentimentIntensityAnalyzer. Default is 'vader_lexicon'.

        Raises:
           None.
        """

        self._scanner = InputSentiment(threshold=threshold, lexicon=lexicon)

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        return self._scanner.scan(output)



"""Input scanners init"""
from .anonymize import Anonymize
from .ban_competitors import BanCompetitors
from .ban_substrings import BanSubstrings
from .ban_topics import BanTopics
from .code import Code
from .invisible_text import InvisibleText
from .language import Language
from .prompt_injection import PromptInjection
from .regex import Regex
from .secrets import Secrets
from .sentiment import Sentiment
from .token_limit import TokenLimit
from .toxicity import Toxicity
from .util import get_scanner_by_name

__all__ = [
    "Anonymize",
    "BanCompetitors",
    "BanSubstrings",
    "BanTopics",
    "Code",
    "InvisibleText",
    "Language",
    "PromptInjection",
    "Regex",
    "Secrets",
    "Sentiment",
    "TokenLimit",
    "Toxicity",
    "get_scanner_by_name",
]



"""
This plugin searches for DigitalOcean tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class DigitaloceanDetector(RegexBasedDetector):
    """Scans for various DigitalOcean Tokens."""

    secret_type = "DigitalOcean Token"

    denylist = [
        # OAuth Access Token
        re.compile(r"""(?i)\b(doo_v1_[a-f0-9]{64})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
        # Personal Access Token
        re.compile(r"""(?i)\b(dop_v1_[a-f0-9]{64})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
        # OAuth Refresh Token
        re.compile(r"""(?i)\b(dor_v1_[a-f0-9]{64})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
    ]



import string
from typing import Optional

from faker import Faker

fake = Faker()
fake.seed_instance(100)

_entity_faker_map = {
    # Global entities
    "CREDIT_CARD": fake.credit_card_number,
    "EMAIL_ADDRESS": fake.email,
    "IBAN_CODE": fake.iban,
    "IP_ADDRESS": fake.ipv4_public,
    "PERSON": fake.name,
    "PHONE_NUMBER": fake.phone_number,
    "URL": fake.url,
    "CREDIT_CARD_RE": fake.credit_card_number,
    "UUID": fake.uuid4,
    "LOCATION": fake.city,
    "DATE_TIME": fake.date,
    "CRYPTO": lambda _: "bc1"
    + "".join(fake.random_choices(string.ascii_lowercase + string.digits, length=26)),
    "NRP": lambda _: str(fake.random_number(digits=8, fix_len=True)),
    "MEDICAL_LICENSE": lambda _: fake.bothify(text="??######").upper(),
    # US-specific entities
    "US_BANK_NUMBER": fake.bban,
    "US_SSN": fake.ssn,
    "US_DRIVER_LICENSE": lambda _: str(fake.random_number(digits=9, fix_len=True)),
    "US_ITIN": lambda _: fake.bothify(text="9##-7#-####"),
    "US_PASSPORT": lambda _: fake.bothify(text="#####??").upper(),
    # UK-specific entities
    "UK_NHS": lambda _: str(fake.random_number(digits=10, fix_len=True)),
    # Spain-specific entities
    "ES_NIF": lambda _: fake.bothify(text="########?").upper(),
    # Italy-specific entities
    "IT_FISCAL_CODE": lambda _: fake.bothify(text="??????##?##?###?").upper(),
    "IT_DRIVER_LICENSE": lambda _: fake.bothify(text="?A#######?").upper(),
    "IT_VAT_CODE": lambda _: fake.bothify(text="IT???????????"),
    "IT_PASSPORT": lambda _: str(fake.random_number(digits=9, fix_len=True)),
    "IT_IDENTITY_CARD": lambda _: lambda _: str(fake.random_number(digits=7, fix_len=True)),
    # Singapore-specific entities
    "SG_NRIC_FIN": lambda _: fake.bothify(text="????####?").upper(),
    # Australia-specific entities
    "AU_ABN": lambda _: str(fake.random_number(digits=11, fix_len=True)),
    "AU_ACN": lambda _: str(fake.random_number(digits=9, fix_len=True)),
    "AU_TFN": lambda _: str(fake.random_number(digits=9, fix_len=True)),
    "AU_MEDICARE": lambda _: str(fake.random_number(digits=10, fix_len=True)),
}


def get_fake_value(entity_type: str) -> Optional[str]:
    if entity_type not in _entity_faker_map:
        return None

    return _entity_faker_map[entity_type]()



"""
This plugin searches for SendGrid API Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class SendGridApiTokenDetector(RegexBasedDetector):
    """Scans for SendGrid API Tokens."""

    secret_type = "SendGrid API Token"

    denylist = [re.compile(r"""(?i)\b(SG\.[a-z0-9=_\-\.]{66})(?:['|\"|\n|\r|\s|\x60|;]|$)""")]



from typing import Dict, Optional, Sequence

from presidio_anonymizer.core.text_replace_builder import TextReplaceBuilder

from llm_guard.exception import LLMGuardValidationError
from llm_guard.util import device, get_logger, lazy_load_dep

from .base import Scanner

LOGGER = get_logger(__name__)

MODEL_BASE = {
    "path": "tomaarsen/span-marker-bert-base-orgs",
}

MODEL_SMALL = {
    "path": "tomaarsen/span-marker-bert-small-orgs",
}

ALL_MODELS = [MODEL_BASE, MODEL_SMALL]


class BanCompetitors(Scanner):
    """
    Scanner to detect competitors in a prompt.

    It uses a named-entity recognition model to extract organizations from the prompt and compares them to a list of competitors.
    """

    def __init__(
        self,
        competitors: Sequence[str],
        *,
        threshold: float = 0.5,
        redact: bool = True,
        model: Optional[Dict] = None,
    ):
        """
        Initialize BanCompetitors object.

        Parameters:
            competitors (Sequence[str]): List of competitors to detect.
            threshold (float, optional): Threshold to determine if a competitor is present in the prompt. Default is 0.5.
            redact (bool, optional): Whether to redact the competitor name. Default is True.
            model (Dict, optional): Model to use for named-entity recognition. Default is BASE model.

        Raises:
            ValueError: If no topics are provided.
        """
        if model is None:
            model = MODEL_BASE

        if model not in ALL_MODELS:
            raise LLMGuardValidationError(f"Model must be in the list of allowed: {ALL_MODELS}")

        self._competitors = competitors
        self._threshold = threshold
        self._redact = redact

        span_marker = lazy_load_dep("span_marker", "span-marker")
        self._ner_pipeline = span_marker.SpanMarkerModel.from_pretrained(
            model["path"], labels=["ORG"]
        )

        if device().type == "cuda":
            self._ner_pipeline = self._ner_pipeline.cuda()

    def scan(self, prompt: str) -> (str, bool, float):
        is_detected = False
        text_replace_builder = TextReplaceBuilder(original_text=prompt)
        entities = self._ner_pipeline.predict(prompt)
        entities = sorted(entities, key=lambda x: x["char_end_index"], reverse=True)
        for entity in entities:
            if entity["span"] not in self._competitors:
                continue

            if entity["score"] < self._threshold:
                LOGGER.debug(
                    "Competitor detected but the score is below threshold",
                    entity=entity["span"],
                    score=entity["score"],
                )
                continue

            is_detected = True

            if self._redact:
                text_replace_builder.replace_text_get_insertion_index(
                    "[REDACTED]",
                    entity["char_start_index"],
                    entity["char_end_index"],
                )

            LOGGER.warning(
                "Competitor detected with score", entity=entity["span"], score=entity["score"]
            )

        if is_detected:
            return text_replace_builder.output_text, False, 1.0

        LOGGER.debug("None of the competitors were detected")

        return prompt, True, 0.0



"""
This plugin searches for EasyPost tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class EasyPostDetector(RegexBasedDetector):
    """Scans for various EasyPost Tokens."""

    secret_type = "EasyPost Token"

    denylist = [
        # EasyPost API token
        re.compile(r"""(?i)\bEZAK[a-z0-9]{54}"""),
        # EasyPost test API token
        re.compile(r"""(?i)\bEZTK[a-z0-9]{54}"""),
    ]



"""
This plugin searches for Doppler API tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class DopplerApiTokenDetector(RegexBasedDetector):
    """Scans for Doppler API Tokens."""

    secret_type = "Doppler API Token"

    denylist = [
        # Doppler API token
        re.compile(r"""(?i)dp\.pt\.[a-z0-9]{43}"""),
    ]



"""
This plugin searches for Alibaba secrets
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class AlibabaSecretDetector(RegexBasedDetector):
    """Scans for Alibaba AccessKey IDs and Secret Keys."""

    secret_type = "Alibaba Secrets"

    denylist = [
        # For Alibaba AccessKey ID
        re.compile(r"""(?i)\b((LTAI)[a-z0-9]{20})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
        # For Alibaba Secret Key
        re.compile(
            r"""(?i)(?:alibaba)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9]{30})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



"""
This plugin searches for Shopify Access Tokens, Custom Access Tokens,
Private App Access Tokens, and Shared Secrets.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class ShopifyDetector(RegexBasedDetector):
    """Scans for Shopify Access Tokens, Custom Access Tokens, Private App Access Tokens,
    and Shared Secrets.
    """

    secret_type = "Shopify Secret"

    denylist = [
        # Shopify access token
        re.compile(r"""shpat_[a-fA-F0-9]{32}"""),
        # Shopify custom access token
        re.compile(r"""shpca_[a-fA-F0-9]{32}"""),
        # Shopify private app access token
        re.compile(r"""shppa_[a-fA-F0-9]{32}"""),
        # Shopify shared secret
        re.compile(r"""shpss_[a-fA-F0-9]{32}"""),
    ]



"""
This plugin searches for Flutterwave API keys.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class FlutterwaveDetector(RegexBasedDetector):
    """Scans for Flutterwave API Keys."""

    secret_type = "Flutterwave API Key"

    denylist = [
        # Flutterwave Encryption Key
        re.compile(r"""(?i)FLWSECK_TEST-[a-h0-9]{12}"""),
        # Flutterwave Public Key
        re.compile(r"""(?i)FLWPUBK_TEST-[a-h0-9]{32}-X"""),
        # Flutterwave Secret Key
        re.compile(r"""(?i)FLWSECK_TEST-[a-h0-9]{32}-X"""),
    ]



import json
import os
import re
from typing import Dict, List, Optional, Sequence

from presidio_anonymizer.core.text_replace_builder import TextReplaceBuilder
from presidio_anonymizer.entities import PIIEntity, RecognizerResult

from ..exception import LLMGuardValidationError
from ..util import get_logger
from ..vault import Vault
from .anonymize_helpers import (
    BERT_BASE_NER_CONF,
    get_analyzer,
    get_fake_value,
    get_transformers_recognizer,
)
from .base import Scanner

LOGGER = get_logger(__name__)

sensitive_patterns_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "..",
    "resources",
    "sensisitive_patterns.json",
)

default_entity_types = [
    "CREDIT_CARD",
    "CRYPTO",
    "EMAIL_ADDRESS",
    "IBAN_CODE",
    "IP_ADDRESS",
    "PERSON",
    "PHONE_NUMBER",
    "US_SSN",
    "US_BANK_NUMBER",
    "CREDIT_CARD_RE",
    "UUID",
    "EMAIL_ADDRESS_RE",
    "US_SSN_RE",
]

ALL_SUPPORTED_LANGUAGES = ["en", "zh"]


class Anonymize(Scanner):
    """
    Anonymize sensitive data in the text using NLP (English only) and predefined regex patterns.

    Anonymizes detected entities with placeholders like [REDACTED_PERSON_1] and stores the real values in a Vault.
    Deanonymizer can be used to replace the placeholders back to their original values.
    """

    def __init__(
        self,
        vault: Vault,
        *,
        hidden_names: Optional[Sequence[str]] = None,
        allowed_names: Optional[Sequence[str]] = None,
        entity_types: Optional[Sequence[str]] = None,
        preamble: str = "",
        regex_pattern_groups_path: str = sensitive_patterns_path,
        use_faker: bool = False,
        recognizer_conf: Optional[Dict] = BERT_BASE_NER_CONF,
        threshold: float = 0,
        use_onnx: bool = False,
        language: str = "en",
    ):
        """
        Initialize an instance of Anonymize class.

        Parameters:
            vault (Vault): A vault instance to store the anonymized data.
            hidden_names (Optional[Sequence[str]]): List of names to be anonymized e.g. [REDACTED_CUSTOM_1].
            allowed_names (Optional[Sequence[str]]): List of names allowed in the text without anonymizing.
            entity_types (Optional[Sequence[str]]): List of entity types to be detected. If not provided, defaults to all.
            preamble (str): Text to prepend to sanitized prompt. If not provided, defaults to an empty string.
            regex_pattern_groups_path (str): Path to a JSON file with regex pattern groups. If not provided, defaults to sensisitive_patterns.json.
            use_faker (bool): Whether to use faker instead of placeholders in applicable cases. If not provided, defaults to False, replaces with placeholders [REDACTED_PERSON_1].
            recognizer_conf (Optional[Dict]): Configuration to recognize PII data. Default is dslim/bert-base-NER.
            threshold (float): Acceptance threshold. Default is 0.
            use_onnx (bool): Whether to use ONNX runtime for inference. Default is False.
            language (str): Language of the anonymize detect. Default is "en".
        """

        if language not in ALL_SUPPORTED_LANGUAGES:
            raise LLMGuardValidationError(
                f"Language must be in the list of allowed: {ALL_SUPPORTED_LANGUAGES}"
            )

        os.environ["TOKENIZERS_PARALLELISM"] = "false"  # Disables huggingface/tokenizers warning

        if not entity_types:
            LOGGER.debug(
                "No entity types provided, using default", default_entities=default_entity_types
            )
            entity_types = default_entity_types.copy()

        entity_types.append("CUSTOM")

        if not hidden_names:
            hidden_names = []

        self._vault = vault
        self._entity_types = entity_types
        self._allowed_names = allowed_names
        self._preamble = preamble
        self._use_faker = use_faker
        self._threshold = threshold
        self._language = language

        transformers_recognizer = get_transformers_recognizer(
            recognizer_conf=recognizer_conf,
            use_onnx=use_onnx,
            supported_language=language,
        )

        self._analyzer = get_analyzer(
            recognizer=transformers_recognizer,
            regex_groups=Anonymize.get_regex_patterns(regex_pattern_groups_path),
            custom_names=hidden_names,
            supported_languages=list(set(["en", language])),
        )

    @staticmethod
    def get_regex_patterns(json_path: str) -> List[dict]:
        """
        Load regex patterns from a specified JSON file.

        Parameters:
            json_path (str): Path to the JSON file containing regex patterns.

        Returns:
            List[dict]: List of regex patterns with each dictionary containing "name", "expressions", "context", and "score".
                        Returns an empty list if file not found or parsing error occurred.
        """
        regex_groups = []
        try:
            with open(json_path, "r") as myfile:
                pattern_groups_raw = json.load(myfile)
            for group in pattern_groups_raw:
                regex_groups.append(
                    {
                        "name": group["name"].upper(),
                        "expressions": group.get("expressions", []),
                        "context": group.get("context", []),
                        "score": group.get("score", 0.75),
                        "languages": group.get("languages", ["en"]),
                        "reuse": group.get("reuse", False),
                    }
                )
                LOGGER.debug("Loaded regex pattern", group_name=group["name"])
        except FileNotFoundError:
            LOGGER.warning("Could not find file", path=json_path)
        except json.decoder.JSONDecodeError as json_error:
            LOGGER.warning("Could not parse file", path=json_path, error=json_error)
        return regex_groups

    @staticmethod
    def _remove_conflicts_and_get_text_manipulation_data(
        analyzer_results: List[RecognizerResult],
    ) -> List[RecognizerResult]:
        """
        Remove conflicting and redundant detections from the analyzer results.
        """
        tmp_analyzer_results = []
        # This list contains all elements which we need to check a single result
        # against. If a result is dropped, it can also be dropped from this list
        # since it is intersecting with another result, and we selected the other one.
        other_elements = analyzer_results.copy()
        for result in analyzer_results:
            other_elements.remove(result)

            is_merge_same_entity_type = False
            for other_element in other_elements:
                if other_element.entity_type != result.entity_type:
                    continue
                if result.intersects(other_element) == 0:
                    continue

                other_element.start = min(result.start, other_element.start)
                other_element.end = max(result.end, other_element.end)
                other_element.score = max(result.score, other_element.score)
                is_merge_same_entity_type = True
                break
            if not is_merge_same_entity_type:
                other_elements.append(result)
                tmp_analyzer_results.append(result)
            else:
                LOGGER.debug("Removing element the results list due to merge", result=result)

        unique_text_metadata_elements = []
        # This list contains all elements which we need to check a single result
        # against. If a result is dropped, it can also be dropped from this list
        # since it is intersecting with another result, and we selected the other one.
        other_elements = tmp_analyzer_results.copy()
        for result in tmp_analyzer_results:
            other_elements.remove(result)
            result_conflicted = Anonymize.__is_result_conflicted_with_other_elements(
                other_elements, result
            )
            if not result_conflicted:
                other_elements.append(result)
                unique_text_metadata_elements.append(result)
            else:
                LOGGER.debug(
                    "Removing element from the results list due to conflict", result=result
                )
        return unique_text_metadata_elements

    @staticmethod
    def __is_result_conflicted_with_other_elements(other_elements, result):
        """
        Check if the given result conflicts with any other elements.
        """
        return any([result.has_conflict(other_element) for other_element in other_elements])

    @staticmethod
    def _merge_entities_with_whitespace_between(
        text: str, analyzer_results: List[RecognizerResult]
    ) -> List[RecognizerResult]:
        """
        Merge adjacent entities of the same type separated by whitespace.
        """
        merged_results = []
        prev_result = None
        for result in analyzer_results:
            if prev_result is not None:
                if prev_result.entity_type == result.entity_type:
                    if re.search(r"^( )+$", text[prev_result.end : result.start]):
                        merged_results.remove(prev_result)
                        result.start = prev_result.start
            merged_results.append(result)
            prev_result = result
        return merged_results

    @staticmethod
    def _get_entity_placeholder(entity_type: str, index: int, use_faker: bool) -> str:
        result = f"[REDACTED_{entity_type}_{index}]"
        if use_faker:
            result = get_fake_value(entity_type) or result

        return result

    @staticmethod
    def _anonymize(
        prompt: str, pii_entities: List[PIIEntity], vault: Vault, use_faker: bool
    ) -> (str, List[tuple]):
        """
        Replace detected entities in the prompt with anonymized placeholders.

        Parameters:
            prompt (str): Original text prompt.
            pii_entities (List[PIIEntity]): List of entities detected in the prompt.
            vault (Vault): A vault instance with the anonymized data stored.
            use_faker (bool): Whether to use faker to generate fake data.

        Returns:
            str: Sanitized text.
            List[tuple]: List of tuples representing the replaced entities and their corresponding placeholders.
        """
        text_replace_builder = TextReplaceBuilder(original_text=prompt)

        entity_type_counter, new_entity_counter = {}, {}
        for pii_entity in pii_entities:
            entity_type = pii_entity.entity_type
            entity_value = text_replace_builder.get_text_in_position(
                pii_entity.start, pii_entity.end
            )

            if entity_type not in entity_type_counter:
                entity_type_counter[entity_type] = {}

            if entity_value not in entity_type_counter[entity_type]:
                vault_entities = [
                    (entity_placeholder, entity_vault_value)
                    for entity_placeholder, entity_vault_value in vault.get()
                    if entity_type in entity_placeholder
                ]
                entity_placeholder = [
                    entity_placeholder
                    for entity_placeholder, entity_vault_value in vault_entities
                    if entity_vault_value == entity_value
                ]
                if len(entity_placeholder) > 0:
                    entity_type_counter[entity_type][entity_value] = int(
                        entity_placeholder[0].split("_")[-1][:-1]
                    )
                else:
                    entity_type_counter[entity_type][entity_value] = (
                        len(vault_entities) + new_entity_counter.get(entity_type, 0) + 1
                    )
                    new_entity_counter[entity_type] = new_entity_counter.get(entity_type, 0) + 1

        results = []
        sorted_pii_entities = sorted(pii_entities, reverse=True)
        for pii_entity in sorted_pii_entities:
            entity_type = pii_entity.entity_type
            entity_value = text_replace_builder.get_text_in_position(
                pii_entity.start, pii_entity.end
            )

            index = entity_type_counter[entity_type][entity_value]
            changed_entity = Anonymize._get_entity_placeholder(entity_type, index, use_faker)

            results.append((changed_entity, prompt[pii_entity.start : pii_entity.end]))

            text_replace_builder.replace_text_get_insertion_index(
                changed_entity, pii_entity.start, pii_entity.end
            )

        return text_replace_builder.output_text, results

    @staticmethod
    def remove_single_quotes(text: str) -> str:
        text_without_single_quotes = text.replace("'", " ")
        return text_without_single_quotes

    def scan(self, prompt: str) -> (str, bool, float):
        risk_score = 0.0
        if prompt.strip() == "":
            return prompt, True, risk_score

        analyzer_results = self._analyzer.analyze(
            text=Anonymize.remove_single_quotes(prompt),
            language=self._language,
            entities=self._entity_types,
            allow_list=self._allowed_names,
            score_threshold=self._threshold,
        )

        risk_score = round(
            max(analyzer_result.score for analyzer_result in analyzer_results)
            if analyzer_results
            else 0.0,
            1,
        )
        analyzer_results = self._remove_conflicts_and_get_text_manipulation_data(analyzer_results)
        merged_results = self._merge_entities_with_whitespace_between(prompt, analyzer_results)

        sanitized_prompt, anonymized_results = self._anonymize(
            prompt, merged_results, self._vault, self._use_faker
        )

        if prompt != sanitized_prompt:
            LOGGER.warning(
                "Found sensitive data in the prompt and replaced it",
                merged_results=merged_results,
                risk_score=risk_score,
            )
            for entity_placeholder, entity_value in anonymized_results:
                if not self._vault.placeholder_exists(entity_placeholder):
                    self._vault.append((entity_placeholder, entity_value))
            return self._preamble + sanitized_prompt, False, risk_score

        LOGGER.debug("Prompt does not have sensitive data to replace", risk_score=risk_score)

        return prompt, True, 0.0



from typing import Dict, Optional

from .ban_competitors import BanCompetitors
from .ban_substrings import BanSubstrings
from .ban_topics import BanTopics
from .base import Scanner
from .bias import Bias
from .code import Code
from .deanonymize import Deanonymize
from .factual_consistency import FactualConsistency
from .json import JSON
from .language import Language
from .language_same import LanguageSame
from .malicious_urls import MaliciousURLs
from .no_refusal import NoRefusal
from .reading_time import ReadingTime
from .regex import Regex
from .relevance import Relevance
from .sensitive import Sensitive
from .sentiment import Sentiment
from .toxicity import Toxicity
from .url_reachabitlity import URLReachability


def get_scanner_by_name(scanner_name: str, scanner_config: Optional[Dict] = None) -> Scanner:
    """
    Get scanner by name.

    Parameters:
        scanner_name (str): Name of scanner.
        scanner_config (Optional[Dict], optional): Scanner configuration. Defaults to None.

    Raises:
        ValueError: If scanner name is unknown.
    """
    if scanner_config is None:
        scanner_config = {}

    if scanner_name == "BanCompetitors":
        return BanCompetitors(**scanner_config)

    if scanner_name == "BanSubstrings":
        return BanSubstrings(**scanner_config)

    if scanner_name == "BanTopics":
        return BanTopics(**scanner_config)

    if scanner_name == "Bias":
        return Bias(**scanner_config)

    if scanner_name == "Deanonymize":
        return Deanonymize(**scanner_config)

    if scanner_name == "FactualConsistency":
        return FactualConsistency(**scanner_config)

    if scanner_name == "JSON":
        return JSON(**scanner_config)

    if scanner_name == "Language":
        return Language(**scanner_config)

    if scanner_name == "LanguageSame":
        return LanguageSame(**scanner_config)

    if scanner_name == "Code":
        return Code(**scanner_config)

    if scanner_name == "MaliciousURLs":
        return MaliciousURLs(**scanner_config)

    if scanner_name == "NoRefusal":
        return NoRefusal(**scanner_config)

    if scanner_name == "ReadingTime":
        return ReadingTime(**scanner_config)

    if scanner_name == "Regex":
        return Regex(**scanner_config)

    if scanner_name == "Relevance":
        return Relevance(**scanner_config)

    if scanner_name == "Sensitive":
        return Sensitive(**scanner_config)

    if scanner_name == "Sentiment":
        return Sentiment(**scanner_config)

    if scanner_name == "Toxicity":
        return Toxicity(**scanner_config)

    if scanner_name == "URLReachability":
        return URLReachability(**scanner_config)

    raise ValueError("Unknown scanner name")



from typing import Dict, Optional, Sequence

from presidio_anonymizer import AnonymizerEngine

from llm_guard.input_scanners.anonymize import (
    Anonymize,
    default_entity_types,
    sensitive_patterns_path,
)
from llm_guard.input_scanners.anonymize_helpers import (
    BERT_BASE_NER_CONF,
    get_analyzer,
    get_transformers_recognizer,
)
from llm_guard.util import get_logger

from .base import Scanner

LOGGER = get_logger(__name__)


class Sensitive(Scanner):
    """
    A class used to detect sensitive (PII) data in the output of a language model.

    This class uses the Presidio Analyzer Engine and predefined internally patterns (sensitive_patterns.json) to analyze the output for specified entity types.
    If no entity types are specified, it defaults to checking for all entity types.
    """

    def __init__(
        self,
        *,
        entity_types: Optional[Sequence[str]] = None,
        regex_pattern_groups_path: str = sensitive_patterns_path,
        redact: bool = False,
        recognizer_conf: Optional[Dict] = BERT_BASE_NER_CONF,
        threshold: float = 0,
        use_onnx: bool = False,
    ):
        """
        Initializes an instance of the Sensitive class.

        Parameters:
           entity_types (Optional[Sequence[str]]): The entity types to look for in the output. Defaults to all
                                               entity types.
           regex_pattern_groups_path (str): Path to the regex patterns file. Default is sensitive_patterns.json.
           redact (bool): Redact found sensitive entities. Default to False.
           recognizer_conf (Optional[Dict]): Configuration to recognize PII data. Default is dslim/bert-base-NER.
           threshold (float): Acceptance threshold. Default is 0.
           use_onnx (bool): Use ONNX model for inference. Default is False.
        """
        if not entity_types:
            LOGGER.debug(
                "No entity types provided, using default", default_entity_types=default_entity_types
            )
            entity_types = default_entity_types.copy()
        entity_types.append("CUSTOM")

        self._entity_types = entity_types
        self._redact = redact
        self._threshold = threshold

        transformers_recognizer = get_transformers_recognizer(recognizer_conf, use_onnx)
        self._analyzer = get_analyzer(
            transformers_recognizer, Anonymize.get_regex_patterns(regex_pattern_groups_path), []
        )
        self._anonymizer = AnonymizerEngine()

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        if output.strip() == "":
            return prompt, True, 0.0

        analyzer_results = self._analyzer.analyze(
            text=Anonymize.remove_single_quotes(output),
            language="en",
            entities=self._entity_types,
            score_threshold=self._threshold,
        )

        if analyzer_results:
            if self._redact:
                LOGGER.debug("Redacting sensitive entities")
                result = self._anonymizer.anonymize(text=output, analyzer_results=analyzer_results)
                output = result.text

            risk_score = max(analyzer_result.score for analyzer_result in analyzer_results)
            LOGGER.warning("Found sensitive data in the output", results=analyzer_results)
            return output, False, risk_score

        LOGGER.debug("No sensitive data found in the output")
        return output, True, 0.0



from typing import Dict, Optional, Sequence

from llm_guard.input_scanners.ban_topics import BanTopics as InputBanTopics

from .base import Scanner


class BanTopics(Scanner):
    """
    A text scanner that checks whether the generated text output includes banned topics.

    The class uses the zero-shot-classification model from Hugging Face to scan the topics present in the text.
    """

    def __init__(
        self,
        topics: Sequence[str],
        *,
        threshold: float = 0.75,
        model: Optional[Dict] = None,
        use_onnx: bool = False,
        transformers_kwargs: Optional[Dict] = None,
    ):
        """
        Initializes BanTopics with a list of topics and a probability threshold.

        Parameters:
            topics (Sequence[str]): The list of topics to be banned from the text.
            threshold (float): The minimum probability required for a topic to be considered present in the text.
                               Default is 0.75.
            model (Dict, optional): The name of the zero-shot-classification model to be used. Default is MODEL_BASE.
            use_onnx (bool, optional): Whether to use ONNX for inference. Default is False.
            transformers_kwargs (Dict, optional): Additional kwargs to pass to the transformers pipeline. Default is None.

        Raises:
            ValueError: If no topics are provided.
        """
        self._scanner = InputBanTopics(
            topics,
            threshold=threshold,
            model=model,
            use_onnx=use_onnx,
            transformers_kwargs=transformers_kwargs,
        )

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        return self._scanner.scan(output)



"""
This plugin searches for Linear API Tokens and Linear Client Secrets.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class LinearDetector(RegexBasedDetector):
    """Scans for Linear secrets."""

    secret_type = "Linear Secret"

    denylist = [
        # Linear API Token
        re.compile(r"""(?i)lin_api_[a-z0-9]{40}"""),
        # Linear Client Secret
        re.compile(
            r"""(?i)(?:linear)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-f0-9]{32})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



import re
from typing import List

from presidio_analyzer import Pattern, PatternRecognizer


class CustomPatternRecognizer(PatternRecognizer):
    def _deny_list_to_regex(self, deny_list: List[str]) -> Pattern:
        """
        Convert a list of characters to a matching regex.

        :param deny_list: the list of characters to detect
        :return:the regex of the characters for detection
        """
        escaped_deny_list = [re.escape(element) for element in deny_list]
        regex = r"(" + "|".join(escaped_deny_list) + r")"
        return Pattern(name="deny_list", regex=regex, score=self.deny_list_score)



import logging
import os
from typing import Sequence, Union

from llm_guard.input_scanners.ban_substrings import BanSubstrings as InputBanSubstrings
from llm_guard.input_scanners.ban_substrings import MatchType

from .base import Scanner

log = logging.getLogger(__name__)
stop_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "..",
    "resources",
    "output_stop_substrings.json",
)


class BanSubstrings(Scanner):
    """
    A text scanner that checks whether the generated text output includes banned substrings.

    The class offers the functionality to match substrings in two different ways: as 'str' or as 'word'.
    - 'str' matches the substring anywhere in the text.
    - 'word' matches the substring as a whole word.
    """

    def __init__(
        self,
        substrings: Sequence[str],
        *,
        match_type: Union[MatchType, str] = MatchType.STR,
        case_sensitive: bool = False,
        redact: bool = False,
        contains_all: bool = False,  # contains any
    ):
        """
        Initializes BanSubstrings with a match type, case sensitivity option, and a list of substrings.

        Parameters:
            substrings (Sequence[str]): The list of substrings to be banned from the text.
            match_type (MatchType): The type of substring matching. Can be either 'str' or 'word'. Default is 'str'.
            case_sensitive (bool): Determines whether the substring matching is case sensitive. Default is False.
            redact (bool): Determines whether the banned substrings should be redacted from the text. Default is False.
            contains_all (bool): Flag to indicate if need to match all substrings instead of any of them. Default is contains any.

        Raises:
            ValueError: If no substrings are provided or match_type is not 'str' or 'word'.
        """

        self._scanner = InputBanSubstrings(
            substrings,
            match_type=match_type,
            case_sensitive=case_sensitive,
            redact=redact,
            contains_all=contains_all,
        )

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        return self._scanner.scan(output)



from typing import Dict, Optional, Sequence

from llm_guard.input_scanners.ban_competitors import BanCompetitors as InputBanCompetitors

from .base import Scanner


class BanCompetitors(Scanner):
    """
    Scanner that detects if an output contains a competitor.

    It uses a named-entity recognition model to extract organization and match it against a list of competitors.
    """

    def __init__(
        self,
        competitors: Sequence[str],
        *,
        threshold: float = 0.5,
        redact: bool = True,
        model: Optional[Dict] = None,
    ):
        """
        Initializes BanCompetitors object.

        Parameters:
            competitors (Sequence[str]): List of competitors to ban.
            threshold (float, optional): Threshold to determine if an organization is present in the output. Default is 0.5.
            redact (bool, optional): Whether to redact the organization name. Default is True.
            model (Dict, optional): Model to use for named-entity recognition. Default is BASE model.

        Raises:
            ValueError: If no competitors are provided.
        """
        self._scanner = InputBanCompetitors(
            competitors,
            threshold=threshold,
            redact=redact,
            model=model,
        )

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        return self._scanner.scan(output)



"""
This plugin searches for Authress Service Client Access Keys
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class AuthressAccessKeyDetector(RegexBasedDetector):
    """Scans for Authress Service Client Access Keys."""

    secret_type = "Authress Service Client Access Key"

    denylist = [
        # For Authress Service Client Access Key
        re.compile(
            r"""(?i)\b((?:sc|ext|scauth|authress)_[a-z0-9]{5,30}\.[a-z0-9]{4,6}\.acc[_-][a-z0-9-]{10,32}\.[a-z0-9+/_=-]{30,120})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



from ..phone_recognizer import PhoneRecognizer
from .crypto_recognizer import CryptoRecognizer
from .custom_pattern_recognizer import CustomPatternRecognizer
from .email_recognizer import EmailRecognizer
from .ip_recognizer import IpRecognizer

__all__ = [
    "PhoneRecognizer",
    "CryptoRecognizer",
    "CustomPatternRecognizer",
    "EmailRecognizer",
    "IpRecognizer",
]



from typing import Dict, Optional

from llm_guard.input_scanners.language import model_path
from llm_guard.transformers_helpers import pipeline
from llm_guard.util import get_logger

from .base import Scanner

LOGGER = get_logger(__name__)


class LanguageSame(Scanner):
    """
    LanguageSame class is responsible for detecting and comparing the language of given prompt and model output to ensure they are the same.
    """

    def __init__(
        self,
        *,
        threshold: float = 0.1,
        use_onnx: bool = False,
        transformers_kwargs: Optional[Dict] = None,
    ):
        """
        Initializes the LanguageSame scanner.

        Parameters:
            threshold (float): Minimum confidence score
            use_onnx (bool): Whether to use ONNX for inference. Default is False.
            transformers_kwargs (dict): Additional keyword arguments to pass to the transformers pipeline.
        """

        self._threshold = threshold

        transformers_kwargs = transformers_kwargs or {}
        transformers_kwargs["max_length"] = 512
        transformers_kwargs["truncation"] = True
        transformers_kwargs["top_k"] = None

        self._pipeline = pipeline(
            task="text-classification",
            model=model_path[0],
            onnx_model=model_path[1],
            use_onnx=use_onnx,
            **transformers_kwargs,
        )

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        if prompt.strip() == "" or output.strip() == "":
            return prompt, True, 0.0

        detected_languages = self._pipeline([prompt, output])
        prompt_languages = [
            detected_language["label"]
            for detected_language in detected_languages[0]
            if detected_language["score"] > self._threshold
        ]
        output_languages = [
            detected_language["label"]
            for detected_language in detected_languages[1]
            if detected_language["score"] > self._threshold
        ]

        if len(prompt_languages) == 0:
            LOGGER.warning("None of languages are above found in the prompt")
            return output, False, 1.0

        if len(output_languages) == 0:
            LOGGER.warning("None of languages are above threshold found in the output")
            return output, False, 1.0

        common_languages = list(set(prompt_languages).intersection(output_languages))
        if len(common_languages) == 0:
            LOGGER.warning(
                "No common languages in the output and prompt", common_languages=common_languages
            )
            return output, False, 1.0

        LOGGER.debug(
            "Languages are found in the prompt and output", common_languages=common_languages
        )
        return output, True, 0.0



"""
This plugin searches for Grafana secrets.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class GrafanaDetector(RegexBasedDetector):
    """Scans for Grafana Secrets."""

    secret_type = "Grafana Secret"

    denylist = [
        # Grafana API key or Grafana Cloud API key
        re.compile(r"""(?i)\b(eyJrIjoi[A-Za-z0-9]{70,400}={0,2})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
        # Grafana Cloud API token
        re.compile(r"""(?i)\b(glc_[A-Za-z0-9+/]{32,400}={0,2})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
        # Grafana Service Account token
        re.compile(r"""(?i)\b(glsa_[A-Za-z0-9]{32}_[A-Fa-f0-9]{8})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
    ]



from enum import Enum
from typing import Dict, List, Optional, Union

from llm_guard.transformers_helpers import pipeline
from llm_guard.util import calculate_risk_score, get_logger, split_text_by_sentences

from .base import Scanner

LOGGER = get_logger(__name__)

_model_path = (
    "unitary/unbiased-toxic-roberta",
    "ProtectAI/unbiased-toxic-roberta-onnx",  # ONNX model
)
_toxic_labels = [
    "toxicity",
    "severe_toxicity",
    "obscene",
    "threat",
    "insult",
    "identity_attack",
    "sexual_explicit",
]


class MatchType(Enum):
    SENTENCE = "sentence"
    FULL = "full"

    def get_inputs(self, prompt: str) -> List[str]:
        if self == MatchType.SENTENCE:
            return split_text_by_sentences(prompt)

        return [prompt]


class Toxicity(Scanner):
    """
    A toxicity scanner that uses a pretrained Hugging Face model to assess the toxicity of a given text.

    If the toxicity score is less than a predefined threshold, the text is considered non-toxic. Otherwise, it is
    considered toxic.
    """

    def __init__(
        self,
        *,
        threshold: float = 0.5,
        match_type: Union[MatchType, str] = MatchType.FULL,
        use_onnx: bool = False,
        transformers_kwargs: Optional[Dict] = None,
    ):
        """
        Initializes Toxicity with a threshold for toxicity.

        Parameters:
           threshold (float): Threshold for toxicity. Default is 0.5.
           match_type (MatchType): Whether to match the full text or individual sentences. Default is MatchType.FULL.
           use_onnx (bool): Whether to use ONNX for inference. Default is False.
           transformers_kwargs (Optional[Dict]): Optional keyword arguments for the transformers pipeline.
        """
        if isinstance(match_type, str):
            match_type = MatchType(match_type)

        self._threshold = threshold
        self._match_type = match_type

        transformers_kwargs = transformers_kwargs or {}
        transformers_kwargs["truncation"] = True
        transformers_kwargs["padding"] = "max_length"
        transformers_kwargs["top_k"] = None
        transformers_kwargs["function_to_apply"] = "sigmoid"

        self._pipeline = pipeline(
            task="text-classification",
            model=_model_path[0],
            onnx_model=_model_path[1],
            use_onnx=use_onnx,
            **transformers_kwargs,
        )

    def scan(self, prompt: str) -> (str, bool, float):
        if prompt.strip() == "":
            return prompt, True, 0.0

        inputs = self._match_type.get_inputs(prompt)

        highest_toxicity_score = 0.0
        toxicity_above_threshold = []
        results_all = self._pipeline(inputs)
        for results_chunk in results_all:
            for result in results_chunk:
                if result["label"] not in _toxic_labels:
                    continue

                if result["score"] > self._threshold:
                    toxicity_above_threshold.append(result)

                if result["score"] > highest_toxicity_score:
                    highest_toxicity_score = result["score"]

        if len(toxicity_above_threshold) > 0:
            LOGGER.warning("Detected toxicity in the text", results=toxicity_above_threshold)

            return prompt, False, calculate_risk_score(highest_toxicity_score, self._threshold)

        LOGGER.debug("Not toxicity found in the text", results=results_all)

        return prompt, True, 0.0



"""
This plugin searches for OpenAI API Keys.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class OpenAIApiKeyDetector(RegexBasedDetector):
    """Scans for OpenAI API Keys."""

    secret_type = "OpenAI API Key"

    denylist = [
        re.compile(
            r"""(?i)\b(sk-[a-zA-Z0-9]{20}T3BlbkFJ[a-zA-Z0-9]{20})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        )
    ]



from enum import Enum
from typing import Dict, List, Optional, Union

from llm_guard.exception import LLMGuardValidationError
from llm_guard.transformers_helpers import pipeline
from llm_guard.util import calculate_risk_score, get_logger, split_text_by_sentences

from .base import Scanner

LOGGER = get_logger(__name__)

# This model is proprietary but open source.
MODEL_LAIYER = {
    "path": "ProtectAI/deberta-v3-base-prompt-injection",
    "onnx_path": "ProtectAI/deberta-v3-base-prompt-injection",  # extract from onnx folder
    "label": "INJECTION",
    "max_length": 512,
}

ALL_MODELS = [
    MODEL_LAIYER,
]


class MatchType(Enum):
    SENTENCE = "sentence"
    FULL = "full"

    def get_inputs(self, prompt: str) -> List[str]:
        if self == MatchType.SENTENCE:
            return split_text_by_sentences(prompt)

        return [prompt]


class PromptInjection(Scanner):
    """
    A prompt injection scanner based on HuggingFace model. It is used to
    detect if a prompt is attempting to perform an injection attack.
    """

    def __init__(
        self,
        *,
        model: Optional[Dict] = None,
        threshold: float = 0.9,
        match_type: Union[MatchType, str] = MatchType.FULL,
        use_onnx: bool = False,
        transformers_kwargs: Optional[Dict] = None,
    ):
        """
        Initializes PromptInjection with a threshold.

        Parameters:
            model (Dict, optional): Chosen model to classify prompt. Default is Laiyer's one.
            threshold (float): Threshold for the injection score. Default is 0.9.
            match_type (MatchType): Whether to match the full text or individual sentences. Default is MatchType.FULL.
            use_onnx (bool): Whether to use ONNX for inference. Defaults to False.
            transformers_kwargs (Optional[Dict]): Optional keyword arguments for the transformers pipeline.

        Raises:
            ValueError: If non-existent models were provided.
        """
        if model is None:
            model = MODEL_LAIYER

        if model not in ALL_MODELS:
            raise LLMGuardValidationError(f"Model must be in the list of allowed: {ALL_MODELS}")

        if isinstance(match_type, str):
            match_type = MatchType(match_type)

        self._threshold = threshold
        self._match_type = match_type
        self._model = model

        transformers_kwargs = transformers_kwargs or {}
        transformers_kwargs["max_length"] = model["max_length"]
        transformers_kwargs["truncation"] = True

        self._pipeline = pipeline(
            task="text-classification",
            model=model["path"],
            use_onnx=use_onnx,
            onnx_model=model["onnx_path"],
            **transformers_kwargs,
        )

    def scan(self, prompt: str) -> (str, bool, float):
        if prompt.strip() == "":
            return prompt, True, 0.0

        highest_score = 0.0
        results_all = self._pipeline(self._match_type.get_inputs(prompt))
        for result in results_all:
            injection_score = round(
                result["score"] if result["label"] == self._model["label"] else 1 - result["score"],
                2,
            )

            if injection_score > highest_score:
                highest_score = injection_score

            if injection_score > self._threshold:
                LOGGER.warning("Detected prompt injection", injection_score=injection_score)

                return prompt, False, calculate_risk_score(injection_score, self._threshold)

        LOGGER.debug("No prompt injection detected", highest_score=highest_score)

        return prompt, True, 0.0



"""
This plugin searches for Telegram Bot API Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class TelegramBotApiTokenDetector(RegexBasedDetector):
    """Scans for Telegram Bot API Tokens."""

    secret_type = "Telegram Bot API Token"

    denylist = [
        re.compile(r"""(?i)(?:^|[^0-9])([0-9]{5,16}:A[a-zA-Z0-9_\-]{34})(?:$|[^a-zA-Z0-9_\-])""")
    ]



from presidio_analyzer import Pattern
from presidio_analyzer.predefined_recognizers import CryptoRecognizer as PresidioCryptoRecognizer


class CryptoRecognizer(PresidioCryptoRecognizer):
    PATTERNS = [
        Pattern("Crypto (Medium)", r"[13][a-km-zA-HJ-NP-Z1-9]{26,33}", 0.5),
    ]



"""
This plugin searches for SendinBlue API Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class SendinBlueApiTokenDetector(RegexBasedDetector):
    """Scans for SendinBlue API Tokens."""

    secret_type = "SendinBlue API Token"

    denylist = [
        re.compile(r"""(?i)\b(xkeysib-[a-f0-9]{64}-[a-z0-9]{16})(?:['|\"|\n|\r|\s|\x60|;]|$)""")
    ]



"""
This plugin searches for HubSpot API Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class HubSpotApiTokenDetector(RegexBasedDetector):
    """Scans for HubSpot API Tokens."""

    secret_type = "HubSpot API Token"

    denylist = [
        # HubSpot API Token
        re.compile(
            r"""(?i)(?:hubspot)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



"""
This plugin searches for Frame.io API tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class FrameIoApiTokenDetector(RegexBasedDetector):
    """Scans for Frame.io API Tokens."""

    secret_type = "Frame.io API Token"

    denylist = [
        # Frame.io API token
        re.compile(r"""(?i)fio-u-[a-z0-9\-_=]{64}"""),
    ]



from typing import List, Optional, Tuple


class Vault:
    """
    This class serves as a vault to store tuples. These tuples are typically placeholder values
    used for anonymization that need to be retained for later decoding.

    The vault provides methods to append single tuples, extend with a list of tuples, remove a tuple,
    and get the list of all stored tuples.
    """

    def __init__(self, tuples: Optional[List[Tuple]] = None):
        if tuples is None:
            tuples = []

        self._tuples = tuples

    def append(self, new_tuple: Tuple):
        self._tuples.append(new_tuple)

    def extend(self, new_tuples: List[Tuple]):
        self._tuples.extend(new_tuples)

    def remove(self, tuple_to_remove: Tuple):
        self._tuples.remove(tuple_to_remove)

    def get(self) -> List[Tuple]:
        return self._tuples

    def placeholder_exists(self, placeholder: str) -> bool:
        for entity_placeholder, _ in self._tuples:
            if placeholder == entity_placeholder:
                return True
        return False



import re
from enum import Enum
from typing import Pattern, Sequence, Union

from presidio_anonymizer.core.text_replace_builder import TextReplaceBuilder

from llm_guard.util import get_logger

from .base import Scanner

LOGGER = get_logger(__name__)


class MatchType(Enum):
    SEARCH = "search"
    FULL_MATCH = "fullmatch"

    def match(self, pattern: Pattern[str], text: str):
        return getattr(pattern, self.value)(text)


class Regex(Scanner):
    """
    A class used to detect patterns in the output of a language model using regular expressions.

    This class relies on the list of regular expressions provided by the user. If any of the patterns
    matches the output, the output is considered invalid. It is also possible to redact the output.
    """

    def __init__(
        self,
        patterns: Sequence[str],
        *,
        is_blocked: bool = True,
        match_type: Union[MatchType, str] = MatchType.SEARCH,
        redact: bool = True,
    ):
        """
        Initializes an instance of the Regex class.

        Parameters:
            patterns (Sequence[str]): A list of regular expressions to use for pattern matching.
            is_blocked (bool): Whether the patterns are blocked or allowed.
            match_type (str): The type of match to use. Can be either "search" or "fullmatch".
            redact (bool): Whether to redact the output or not.

        Raises:
            ValueError: If no patterns provided or both good and bad patterns provided.
        """
        if isinstance(match_type, str):
            match_type = MatchType(match_type)

        self._patterns = []
        for pattern in patterns:
            self._patterns.append(re.compile(pattern))

        self._match_type = match_type
        self._is_blocked = is_blocked
        self._redact = redact

    def scan(self, prompt: str) -> (str, bool, float):
        text_replace_builder = TextReplaceBuilder(original_text=prompt)
        for pattern in self._patterns:
            match = self._match_type.match(pattern, prompt)
            if match is None:
                continue

            if self._is_blocked:
                LOGGER.warning("Pattern was detected in the text", pattern=pattern)

                if self._redact:
                    text_replace_builder.replace_text_get_insertion_index(
                        "[REDACTED]",
                        match.start(),
                        match.end(),
                    )

                return text_replace_builder.output_text, False, 1.0

            LOGGER.debug("Pattern matched the text", pattern=pattern)
            return text_replace_builder.output_text, True, 0.0

        if self._is_blocked:
            LOGGER.debug("None of the patterns were found in the text")
            return text_replace_builder.output_text, True, 0.0

        LOGGER.warning("None of the patterns matched the text")
        return text_replace_builder.output_text, False, 1.0



from presidio_analyzer import Pattern
from presidio_analyzer.predefined_recognizers import EmailRecognizer as PresidioEmailRecognizer


class EmailRecognizer(PresidioEmailRecognizer):
    PATTERNS = [
        Pattern(
            "Email (Medium)",
            r"[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Za-z]{2,}",  # noqa: E501
            0.5,
        ),
    ]



from typing import Dict, Optional, Sequence, Union

from llm_guard.input_scanners.language import Language as InputLanguage
from llm_guard.input_scanners.language import MatchType

from .base import Scanner


class Language(Scanner):
    """
    Language scanner is responsible for determining the language of a given text
    prompt and verifying its validity against a list of predefined languages.
    """

    def __init__(
        self,
        valid_languages: Sequence[str],
        *,
        threshold: float = 0.7,
        match_type: Union[MatchType, str] = MatchType.FULL,
        use_onnx: bool = False,
        transformers_kwargs: Optional[Dict] = None,
    ):
        """
        Initializes the Language scanner with a list of valid languages.

        Parameters:
            valid_languages (Sequence[str]): A list of valid language codes.
            threshold (float): Minimum confidence score.
            match_type (MatchType): Whether to match the full text or individual sentences. Default is MatchType.FULL.
            use_onnx (bool): Whether to use ONNX for inference. Default is False.
            transformers_kwargs (dict): Additional keyword arguments to pass to the transformers pipeline.
        """

        self._scanner = InputLanguage(
            valid_languages,
            threshold=threshold,
            match_type=match_type,
            use_onnx=use_onnx,
            transformers_kwargs=transformers_kwargs,
        )

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        return self._scanner.scan(output)



"""
This plugin searches for Scalingo API Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class ScalingoApiTokenDetector(RegexBasedDetector):
    """Scans for Scalingo API Tokens."""

    secret_type = "Scalingo API Token"

    denylist = [re.compile(r"""\btk-us-[a-zA-Z0-9-_]{48}\b""")]



from presidio_analyzer.predefined_recognizers import PhoneRecognizer as PresidioPhoneRecognizer


class PhoneRecognizer(PresidioPhoneRecognizer):
    DEFAULT_SUPPORTED_REGIONS = ("US", "UK", "DE", "FE", "IL", "IN", "CA", "BR", "CN")



"""
This plugin searches for GitLab secrets.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class GitLabDetector(RegexBasedDetector):
    """Scans for GitLab Secrets."""

    secret_type = "GitLab Secret"

    denylist = [
        # GitLab Personal Access Token
        re.compile(r"""glpat-[0-9a-zA-Z\-\_]{20}"""),
        # GitLab Pipeline Trigger Token
        re.compile(r"""glptt-[0-9a-f]{40}"""),
        # GitLab Runner Registration Token
        re.compile(r"""GR1348941[0-9a-zA-Z\-\_]{20}"""),
    ]



import json
import re

import regex

from llm_guard.util import get_logger, lazy_load_dep

from .base import Scanner

LOGGER = get_logger(__name__)
JSON_PATTERN = r"(?<!\\)(?:\\\\)*\{(?:[^{}]|(?R))*\}"


class JSON(Scanner):
    """
    A scanner class to detect, validate and repair JSON structures within a given output.

    It primarily serves to detect JSON objects and arrays using regular expressions,
    then to validate them to ensure their correctness and finally to repair them if necessary.
    """

    def __init__(self, *, required_elements: int = 0, repair: bool = True):
        """Initialize the JSON scanner.

        Parameters:
            required_elements (int, optional): The minimum number of JSON elements
            that should be present. Defaults to 0.
            repair (bool, optional): Whether to repair the broken JSON. Defaults to False.
        """
        self._required_elements = required_elements
        self._repair = repair
        self._pattern = regex.compile(JSON_PATTERN, re.DOTALL)

    @staticmethod
    def is_valid_json(json_str: str) -> bool:
        """Check if the input string is a valid JSON.

        Parameters:
            json_str (str): The input string to check.

        Returns:
            bool: True if the input string is a valid JSON, False otherwise.
        """
        try:
            json.loads(json_str)
            return True
        except ValueError as e:
            LOGGER.warning("Invalid JSON", error=e)
            return False

    @staticmethod
    def repair_json(json_str: str) -> str:
        """Repair a broken JSON string.

        Parameters:
            json_str (str): The input string to repair.

        Returns:
            str: The repaired JSON string.
        """

        json_repair = lazy_load_dep("json_repair")
        try:
            repaired_json = json_repair.repair_json(
                json_str, skip_json_loads=True, return_objects=False
            )
        except ValueError:
            return json_str

        return repaired_json

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        if prompt.strip() == "":
            return output, True, 0.0

        # Find JSON object and array candidates using regular expressions
        json_candidates = self._pattern.findall(output)

        # Validate each JSON
        valid_jsons = []
        for json_candidate in json_candidates:
            if self.is_valid_json(json_candidate):
                valid_jsons.append(json_candidate)
                continue

            if self._repair:
                LOGGER.warning(
                    "Found invalid JSON. Trying to repair it...", json_candidate=json_candidate
                )

                repaired_json = self.repair_json(json_candidate)
                if not self.is_valid_json(repaired_json):
                    LOGGER.warning(
                        "Could not repair JSON. Skipping...", repaired_json=repaired_json
                    )

                    continue

                valid_jsons.append(repaired_json)
                output = output.replace(json_candidate, repaired_json)

        # Check if there are enough valid JSONs
        if len(valid_jsons) < self._required_elements:
            LOGGER.warning(
                "Not all required JSON elements are found",
                num_required_elements=self._required_elements,
                num_found=len(valid_jsons),
            )
            return output, False, 1.0

        # Compare
        if len(valid_jsons) != len(json_candidates):
            LOGGER.warning(
                "Only some JSONs are valid",
                num_valid=len(valid_jsons),
                num_total=len(json_candidates),
            )

            return output, False, 1.0

        return output, True, 0.0



"""
This plugin searches for Defined Networking API Tokens.
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class DefinedNetworkingApiTokenDetector(RegexBasedDetector):
    """Scans for Defined Networking API Tokens."""

    secret_type = "Defined Networking API Token"

    denylist = [
        re.compile(
            r"""(?i)(?:dnkey)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}(dnkey-[a-z0-9=_\-]{26}-[a-z0-9=_\-]{52})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



"""
This plugin searches for GoCardless API tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class GoCardlessApiTokenDetector(RegexBasedDetector):
    """Scans for GoCardless API Tokens."""

    secret_type = "GoCardless API Token"

    denylist = [
        # GoCardless API token
        re.compile(
            r"""(?:gocardless)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}(live_[a-z0-9\-_=]{40})(?:['|\"|\n|\r|\s|\x60|;]|$)""",
            re.IGNORECASE,
        ),
    ]



from typing import Dict, Optional, Union

from llm_guard.input_scanners.toxicity import MatchType
from llm_guard.input_scanners.toxicity import Toxicity as InputToxicity

from .base import Scanner


class Toxicity(Scanner):
    """
    A class used to detect toxicity in the output of a language model.

    This class uses a pre-trained toxicity model from HuggingFace to calculate a toxicity score for the output.
    """

    def __init__(
        self,
        *,
        threshold: float = 0.7,
        match_type: Union[MatchType, str] = MatchType.FULL,
        use_onnx: bool = False,
        transformers_kwargs: Optional[Dict] = None,
    ):
        """
        Initializes an instance of the Toxicity class.

        Parameters:
            threshold (float): The threshold used to determine toxicity. Defaults to 0.7.
            match_type (MatchType): Whether to match the full text or individual sentences. Defaults to MatchType.FULL.
            use_onnx (bool): Whether to use ONNX for inference. Defaults to False.
            transformers_kwargs (dict): Additional keyword arguments to pass to the transformers pipeline.
        """

        self._scanner = InputToxicity(
            threshold=threshold,
            match_type=match_type,
            use_onnx=use_onnx,
            transformers_kwargs=transformers_kwargs,
        )

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        return self._scanner.scan(output)



from typing import Dict, Optional

from .anonymize import Anonymize
from .ban_competitors import BanCompetitors
from .ban_substrings import BanSubstrings
from .ban_topics import BanTopics
from .base import Scanner
from .code import Code
from .invisible_text import InvisibleText
from .language import Language
from .prompt_injection import PromptInjection
from .regex import Regex
from .secrets import Secrets
from .sentiment import Sentiment
from .token_limit import TokenLimit
from .toxicity import Toxicity


def get_scanner_by_name(scanner_name: str, scanner_config: Optional[Dict] = None) -> Scanner:
    """
    Returns a scanner by name.

    Parameters:
        scanner_name (str): The name of the scanner.
        scanner_config (Optional[Dict], optional): The configuration for the scanner. Defaults to None.

    Raises:
        ValueError: If the scanner name is unknown.
    """
    if scanner_config is None:
        scanner_config = {}

    if scanner_name == "Anonymize":
        return Anonymize(**scanner_config)

    if scanner_name == "BanCompetitors":
        return BanCompetitors(**scanner_config)

    if scanner_name == "BanSubstrings":
        return BanSubstrings(**scanner_config)

    if scanner_name == "BanTopics":
        return BanTopics(**scanner_config)

    if scanner_name == "Code":
        return Code(**scanner_config)

    if scanner_name == "InvisibleText":
        return InvisibleText()

    if scanner_name == "Language":
        return Language(**scanner_config)

    if scanner_name == "PromptInjection":
        return PromptInjection(**scanner_config)

    if scanner_name == "Regex":
        return Regex(**scanner_config)

    if scanner_name == "Secrets":
        return Secrets(**scanner_config)

    if scanner_name == "Sentiment":
        return Sentiment(**scanner_config)

    if scanner_name == "TokenLimit":
        return TokenLimit(**scanner_config)

    if scanner_name == "Toxicity":
        return Toxicity(**scanner_config)

    raise ValueError("Unknown scanner name")



import copy
from typing import Dict, List, Sequence

import spacy
from presidio_analyzer import (
    AnalyzerEngine,
    EntityRecognizer,
    Pattern,
    PatternRecognizer,
    RecognizerRegistry,
)
from presidio_analyzer.nlp_engine import NlpEngine, NlpEngineProvider

from llm_guard.exception import LLMGuardValidationError

from .ner_mapping import ALL_RECOGNIZER_CONF
from .predefined_recognizers import _get_predefined_recognizers
from .predefined_recognizers.zh import CustomPatternRecognizer
from .transformers_recognizer import TransformersRecognizer


def _add_recognizers(
    registry: RecognizerRegistry,
    regex_groups,
    custom_names,
    supported_languages: List[str] = ["en"],
) -> RecognizerRegistry:
    """
    Create a RecognizerRegistry and populate it with regex patterns and custom names.

    Parameters:
        regex_groups: List of regex patterns.
        custom_names: List of custom names to recognize.

    Returns:
        RecognizerRegistry: A RecognizerRegistry object loaded with regex and custom name recognizers.
    """

    for language in supported_languages:
        # custom recognizer per language
        if len(custom_names) > 0:
            custom_recognier = PatternRecognizer

            if language == "zh":
                custom_recognier = CustomPatternRecognizer

            registry.add_recognizer(
                custom_recognier(
                    supported_entity="CUSTOM",
                    supported_language=language,
                    deny_list=custom_names,
                )
            )

        # predefined recognizers per language
        for _Recognizer in _get_predefined_recognizers(language):
            registry.add_recognizer(_Recognizer(supported_language=language))

    for pattern_data in regex_groups:
        languages = pattern_data["languages"] or ["en"]

        label = pattern_data["name"]
        reuse = pattern_data.get("reuse", False)

        patterns = map(
            lambda exp: Pattern(name=label, regex=exp, score=pattern_data["score"]),
            pattern_data.get("expressions", []) or [],
        )

        for language in languages:
            if language not in supported_languages:
                continue

            if reuse:
                new_recognizer = copy.deepcopy(
                    registry.get_recognizers(language=reuse["language"], entities=[reuse["name"]])[
                        0
                    ]
                )
                new_recognizer.supported_language = language
                registry.add_recognizer(new_recognizer)
            else:
                registry.add_recognizer(
                    PatternRecognizer(
                        supported_entity=label,
                        supported_language=language,
                        patterns=patterns,
                        context=pattern_data["context"],
                    )
                )

    return registry


def _get_nlp_engine(languages: List[str] = ["en"]) -> NlpEngine:
    models = []

    for language in languages:
        if not spacy.util.is_package(f"{language}_core_web_sm"):
            # Use small spacy model, for faster inference.
            spacy.cli.download(f"{language}_core_web_sm")
        models.append({"lang_code": language, "model_name": f"{language}_core_web_sm"})

    configuration = {"nlp_engine_name": "spacy", "models": models}

    return NlpEngineProvider(nlp_configuration=configuration).create_engine()


def get_transformers_recognizer(
    recognizer_conf: Dict, use_onnx: bool = False, supported_language: str = "en"
) -> EntityRecognizer:
    if recognizer_conf not in ALL_RECOGNIZER_CONF:
        raise LLMGuardValidationError(
            f"Recognizer must be in the list of allowed: {ALL_RECOGNIZER_CONF}"
        )

    model_path = recognizer_conf.get("DEFAULT_MODEL_PATH")
    supported_entities = recognizer_conf.get("PRESIDIO_SUPPORTED_ENTITIES")
    transformers_recognizer = TransformersRecognizer(
        model_path=model_path,
        supported_entities=supported_entities,
        supported_language=supported_language,
    )
    transformers_recognizer.load_transformer(use_onnx=use_onnx, **recognizer_conf)
    return transformers_recognizer


def get_analyzer(
    recognizer: EntityRecognizer,
    regex_groups,
    custom_names: Sequence[str],
    supported_languages: List[str] = ["en"],
) -> AnalyzerEngine:
    nlp_engine = _get_nlp_engine(languages=supported_languages)

    registry = RecognizerRegistry()
    registry.load_predefined_recognizers(nlp_engine=nlp_engine)
    registry = _add_recognizers(registry, regex_groups, custom_names, supported_languages)
    registry.add_recognizer(recognizer)
    registry.remove_recognizer("SpacyRecognizer")

    return AnalyzerEngine(
        nlp_engine=nlp_engine,
        registry=registry,
        supported_languages=supported_languages,
    )



"""
This plugin searches for Facebook Access Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class FacebookAccessTokenDetector(RegexBasedDetector):
    """Scans for Facebook Access Tokens."""

    secret_type = "Facebook Access Token"

    denylist = [
        # Facebook Access Token
        re.compile(
            r"""(?i)(?:facebook)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-f0-9]{32})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



from enum import Enum
from typing import Dict, List, Optional, Union

from llm_guard.transformers_helpers import pipeline
from llm_guard.util import calculate_risk_score, get_logger, split_text_by_sentences

from .base import Scanner

LOGGER = get_logger(__name__)

_model_path = (
    "valurank/distilroberta-bias",
    "ProtectAI/distilroberta-bias-onnx",  # ONNX model
)


class MatchType(Enum):
    SENTENCE = "sentence"
    FULL = "full"

    def get_inputs(self, prompt: str) -> List[str]:
        if self == MatchType.SENTENCE:
            return split_text_by_sentences(prompt)

        return [prompt]


class Bias(Scanner):
    """
    This class is designed to detect and evaluate potential biases in text using a pretrained model from HuggingFace.
    """

    def __init__(
        self,
        *,
        threshold: float = 0.7,
        match_type: Union[MatchType, str] = MatchType.FULL,
        use_onnx: bool = False,
        transformers_kwargs: Optional[Dict] = None,
    ):
        """
        Initializes the Bias scanner with a probability threshold for bias detection.

        Parameters:
           threshold (float): The threshold above which a text is considered biased. Default is 0.7.
           match_type (MatchType): Whether to match the full text or individual sentences. Default is MatchType.FULL.
           use_onnx (bool): Whether to use ONNX instead of PyTorch for inference.
           transformers_kwargs (dict): Additional keyword arguments to pass to the transformers pipeline.
        """
        if isinstance(match_type, str):
            match_type = MatchType(match_type)

        self._threshold = threshold
        self._match_type = match_type

        transformers_kwargs = transformers_kwargs or {}
        transformers_kwargs["truncation"] = True

        self._classifier = pipeline(
            task="text-classification",
            model=_model_path[0],
            onnx_model=_model_path[1],
            use_onnx=use_onnx,
            **transformers_kwargs,
        )

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        if output.strip() == "":
            return output, True, 0.0

        highest_score = 0.0
        results_all = self._classifier(self._match_type.get_inputs(output))
        for result in results_all:
            score = round(
                result["score"] if result["label"] == "BIASED" else 1 - result["score"],
                2,
            )

            if score > highest_score:
                highest_score = score

            if score > self._threshold:
                LOGGER.warning("Detected biased text", score=score, threshold=self._threshold)

                return output, False, calculate_risk_score(score, self._threshold)

        LOGGER.debug("Not biased result", highest_score=highest_score, threshold=self._threshold)

        return output, True, 0.0



"""
This plugin searches for GCP API keys.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class GCPApiKeyDetector(RegexBasedDetector):
    """Scans for GCP API keys."""

    secret_type = "GCP API Key"

    denylist = [
        # GCP API Key
        re.compile(r"""(?i)\b(AIza[0-9A-Za-z\\-_]{35})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
    ]



from typing import Dict, Optional, Sequence

from llm_guard.input_scanners.code import Code as InputCode

from .base import Scanner


class Code(Scanner):
    """
    A class for scanning if the model output includes code in specific programming languages.

    This class uses the transformers library to detect code snippets in the output of the language model.
    It can be configured to allow or deny specific programming languages.
    """

    def __init__(
        self,
        languages: Sequence[str],
        *,
        is_blocked: bool = True,
        threshold: float = 0.5,
        use_onnx: bool = False,
        transformers_kwargs: Optional[Dict] = None,
    ):
        """
        Initializes an instance of the Code class.

        Parameters:
            languages (Sequence[str]): The list of programming languages to allow or deny.
            is_blocked (bool): Whether the languages are blocked or allowed. Default is True.
            threshold (float): The threshold for the model output to be considered valid. Default is 0.5.
            use_onnx (bool): Whether to use ONNX for inference. Default is False.
            transformers_kwargs: Optional[Dict] = None,

        Raises:
            ValueError: If both 'allowed' and 'denied' lists are provided or if both are empty.
        """

        self._scanner = InputCode(
            languages,
            is_blocked=is_blocked,
            threshold=threshold,
            use_onnx=use_onnx,
            transformers_kwargs=transformers_kwargs,
        )

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        return self._scanner.scan(output)



from typing import List

from presidio_analyzer import EntityRecognizer


def _get_predefined_recognizers(language: str) -> List[EntityRecognizer]:
    if language == "zh":
        from .zh import CryptoRecognizer, EmailRecognizer, IpRecognizer, PhoneRecognizer

        return [
            CryptoRecognizer,
            PhoneRecognizer,
            EmailRecognizer,
            IpRecognizer,
        ]
    else:
        return []


__all__ = [
    "_get_predefined_recognizers",
]



from typing import Dict, Optional

from llm_guard.transformers_helpers import pipeline
from llm_guard.util import calculate_risk_score, extract_urls, get_logger

from .base import Scanner

LOGGER = get_logger(__name__)
_model_path = (
    "DunnBC22/codebert-base-Malicious_URLs",
    "ProtectAI/codebert-base-Malicious_URLs-onnx",  # ONNX version
)

_malicious_labels = [
    "defacement",
    "phishing",
    "malware",
]


class MaliciousURLs(Scanner):
    """
    This scanner is used to scan and detect malicious URLs in the text.

    Large Language Models can occasionally produce malicious URLs in their outputs.
    Using the "DunnBC22/codebert-base-Malicious_URLs" model from HuggingFace, this class classifies
    URLs as either malicious or benign to safeguard users from potential threats.
    """

    def __init__(
        self, *, threshold=0.5, use_onnx: bool = False, transformers_kwargs: Optional[Dict] = None
    ):
        """
        Initializes an instance of the MaliciousURLs class.

        Parameters:
            threshold (float): The threshold used to determine if the website is malicious. Defaults to 0.5.
            use_onnx (bool): Whether to use the ONNX version of the model. Defaults to False.
            transformers_kwargs (dict): Additional keyword arguments to pass to the transformers pipeline.
        """

        self._threshold = threshold

        transformers_kwargs = transformers_kwargs or {}
        transformers_kwargs["max_length"] = 512
        transformers_kwargs["truncation"] = True
        transformers_kwargs["top_k"] = None

        self._classifier = pipeline(
            task="text-classification",
            model=_model_path[0],
            onnx_model=_model_path[1],
            use_onnx=use_onnx,
            **transformers_kwargs,
        )

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        if output.strip() == "":
            return output, True, 0.0

        urls = extract_urls(output)
        if len(urls) == 0:
            return output, True, 0.0

        LOGGER.debug("Found URLs in the output", len=len(urls))

        results = self._classifier(urls)
        for url, result in zip(urls, results):
            malicious_scores = [
                item["score"] for item in result if item["label"] in _malicious_labels
            ]
            highest_malicious_score = max(malicious_scores)
            if highest_malicious_score > self._threshold:
                LOGGER.warning(
                    "Detected malware URL",
                    url=url,
                    highest_malicious_score=highest_malicious_score,
                )

                return output, False, calculate_risk_score(highest_malicious_score, self._threshold)

        LOGGER.debug("Not malware URLs in the output", results=results)

        return output, True, 0.0



"""
This plugin searches for PlanetScale API tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class PlanetScaleDetector(RegexBasedDetector):
    """Scans for PlanetScale API Tokens."""

    secret_type = "PlanetScale API Token"

    denylist = [
        # the PlanetScale API token
        re.compile(r"""(?i)\b(pscale_tkn_[a-z0-9=\-_\.]{32,64})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
        # the PlanetScale OAuth token
        re.compile(r"""(?i)\b(pscale_oauth_[a-z0-9=\-_\.]{32,64})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
        # the PlanetScale password
        re.compile(r"""(?i)\b(pscale_pw_[a-z0-9=\-_\.]{32,64})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
    ]



"""
This plugin searches for JFrog-related secrets like API Key and Identity Token.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class JFrogDetector(RegexBasedDetector):
    """Scans for JFrog-related secrets."""

    secret_type = "JFrog Secrets"

    denylist = [
        # JFrog API Key
        re.compile(
            r"""(?i)(?:jfrog|artifactory|bintray|xray)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9]{73})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
        # JFrog Identity Token
        re.compile(
            r"""(?i)(?:jfrog|artifactory|bintray|xray)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9]{64})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



"""
This plugin searches for Sidekiq secrets and sensitive URLs.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class SidekiqDetector(RegexBasedDetector):
    """Scans for Sidekiq secrets and sensitive URLs."""

    secret_type = "Sidekiq Secret"

    denylist = [
        # Sidekiq Secret
        re.compile(
            r"""(?i)(?:BUNDLE_ENTERPRISE__CONTRIBSYS__COM|BUNDLE_GEMS__CONTRIBSYS__COM)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-f0-9]{8}:[a-f0-9]{8})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
        # Sidekiq Sensitive URL
        re.compile(
            r"""(?i)\b(http(?:s??):\/\/)([a-f0-9]{8}:[a-f0-9]{8})@(?:gems.contribsys.com|enterprise.contribsys.com)(?:[\/|\#|\?|:]|$)"""
        ),
    ]



"""
This plugin searches for Beamer API tokens
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class BeamerApiTokenDetector(RegexBasedDetector):
    """Scans for Beamer API tokens."""

    secret_type = "Beamer API token"

    denylist = [
        # For Beamer API token
        re.compile(
            r"""(?i)(?:beamer)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}(b_[a-z0-9=_\-]{44})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



import re

from detect_secrets.plugins.base import RegexBasedDetector


class VaultDetector(RegexBasedDetector):
    """Scans for Vault Batch Tokens and Vault Service Tokens."""

    secret_type = "Vault Token"

    denylist = [
        # Vault Batch Token
        re.compile(r"""(?i)\b(hvb\.[a-z0-9_-]{138,212})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
        # Vault Service Token
        re.compile(r"""(?i)\b(hvs\.[a-z0-9_-]{90,100})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
    ]



"""
This plugin searches for Contentful delivery API token.
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class ContentfulApiTokenDetector(RegexBasedDetector):
    """Scans for Contentful delivery API token."""

    secret_type = "Contentful API Token"

    denylist = [
        re.compile(
            r"""(?i)(?:contentful)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9=_\-]{43})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



"""
This plugin searches for Etsy Access Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class EtsyAccessTokenDetector(RegexBasedDetector):
    """Scans for Etsy Access Tokens."""

    secret_type = "Etsy Access Token"

    denylist = [
        # Etsy Access Token
        re.compile(
            r"""(?i)(?:etsy)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9]{24})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



"""
This plugin searches for Confluent Access Token and Confluent Secret Key
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class ConfluentDetector(RegexBasedDetector):
    """Scans for Confluent Access Token and Confluent Secret Key."""

    secret_type = "Confluent Secret"

    denylist = [
        # For Confluent Access Token
        re.compile(
            r"""(?i)(?:confluent)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9]{16})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
        # For Confluent Secret Key
        re.compile(
            r"""(?i)(?:confluent)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9]{64})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



import os
import re
from enum import Enum
from typing import Sequence, Union

from llm_guard.util import get_logger

from .base import Scanner

LOGGER = get_logger(__name__)

stop_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "..",
    "resources",
    "prompt_stop_substrings.json",
)


class MatchType(Enum):
    STR = "str"
    WORD = "word"

    def match(self, text: str, substring: str) -> bool:
        if self == MatchType.STR:
            return substring in text

        if self == MatchType.WORD:
            return re.search(r"\b" + substring + r"\b", text) is not None


class BanSubstrings(Scanner):
    """
    BanSubstrings class is used to ban certain substrings from appearing in the prompt.

    The match can be done either at a string level or word level.
    """

    def __init__(
        self,
        substrings: Sequence[str],
        *,
        match_type: Union[MatchType, str] = MatchType.STR,
        case_sensitive: bool = False,
        redact: bool = False,
        contains_all: bool = False,  # contains any
    ):
        """
        Initialize BanSubstrings object.

        Parameters:
            substrings (Sequence[str]): List of substrings to ban.
            match_type (MatchType): Type of match to perform. Can be either 'str' or 'word'. Default is 'str'.
            case_sensitive (bool, optional): Flag to indicate if the match should be case-sensitive. Default is False.
            redact (bool, optional): Flag to indicate if the banned substrings should be redacted. Default is False.
            contains_all (bool): Flag to indicate if need to match all substrings instead of any of them. Default is contains any.

        Raises:
            ValueError: If no substrings are provided or match_type is not 'str' or 'word'.
        """
        if isinstance(match_type, str):
            match_type = MatchType(match_type)

        self._match_type = match_type
        self._case_sensitive = case_sensitive
        self._substrings = substrings
        self._redact = redact
        self._contains_all = contains_all

    @staticmethod
    def _redact_text(text: str, substrings: Sequence[str]) -> str:
        redacted_text = text
        for s in substrings:
            redacted_text = redacted_text.replace(s, "[REDACTED]")
        return redacted_text

    def scan(self, prompt: str) -> (str, bool, float):
        sanitized_prompt = prompt
        matched_substrings = []
        missing_substrings = []

        for s in self._substrings:
            if self._case_sensitive is False:
                s, prompt = s.lower(), prompt.lower()

            if self._match_type.match(prompt, s):
                matched_substrings.append(s)
            else:
                missing_substrings.append(s)

        if self._contains_all:
            if len(missing_substrings) > 0:
                LOGGER.debug(
                    "Some substrings were not found", missing_substrings=missing_substrings
                )
                return sanitized_prompt, True, 0.0

            if self._redact:
                sanitized_prompt = self._redact_text(sanitized_prompt, matched_substrings)
                LOGGER.debug("Redacted banned substrings")

            LOGGER.warning("All substrings were found")

            return sanitized_prompt, False, 1.0

        if matched_substrings:
            LOGGER.warning(
                "Found the following banned substrings", matched_substrings=matched_substrings
            )

            if self._redact:
                sanitized_prompt = self._redact_text(sanitized_prompt, matched_substrings)
                LOGGER.debug("Redacted banned substrings")

            return sanitized_prompt, False, 1.0

        LOGGER.debug("No banned substrings found")

        return sanitized_prompt, True, 0.0



from typing import Protocol


class Scanner(Protocol):
    """
    An interface for text scanners.

    This protocol defines the contract for classes that scan text outputs from a language model.
    """

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        """
        Analyzes output of the model and returns sanitized output with a flag indicating if it is valid or malicious.

        Parameters:
            prompt (str): The input prompt.
            output (str): The text output from the language model.

        Returns:
            str: The sanitized and processed output as per the scanner's implementation.
            bool: A flag indicating whether the output is valid or not.
            float: Risk score where 0 means no risk and 1 means high risk.
        """



from llm_guard.util import get_logger

from .base import Scanner

LOGGER = get_logger(__name__)


class ReadingTime(Scanner):
    """
    Scanner that checks the reading time of the output against a maximum time.

    If the output exceeds the maximum time, the output will be truncated to fit within the time limit.
    """

    def __init__(self, max_time: float, *, truncate: bool = False):
        """
        Parameters:
            max_time: Maximum time in minutes that the user should spend reading the output.
            truncate: If True, the output will be truncated to the maximum time.
        """
        self._max_time = max_time
        self._truncate = truncate
        self._words_per_minute = 200  # Average reading speed

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        words = output.split()
        word_count = len(words)
        reading_time_minutes = word_count / self._words_per_minute

        if reading_time_minutes > self._max_time:
            LOGGER.warning(
                "Output exceeded maximum reading time",
                reading_time_minutes=reading_time_minutes,
                max_reading_time_minutes=self._max_time,
            )
            if self._truncate:
                # Calculate the maximum number of words to fit within the time limit
                max_words = int(self._max_time * self._words_per_minute)
                output = " ".join(words[:max_words])

            return output, False, 1.0

        LOGGER.debug("Output reading time", reading_time_minutes=reading_time_minutes)
        return output, True, 0.0



from typing import List

import requests

from llm_guard.util import extract_urls, get_logger

from .base import Scanner

LOGGER = get_logger(__name__)


class URLReachability(Scanner):
    """
    This scanner checks URLs for their reachability.
    """

    def __init__(self, *, success_status_codes: List[int] = None, timeout: int = 5):
        """
        Parameters:
            success_status_codes: A list of status codes that are considered as successful.
            timeout: The timeout in seconds for the HTTP requests.
        """
        if success_status_codes is None:
            success_status_codes = [
                requests.codes.ok,
                requests.codes.created,
                requests.codes.accepted,
            ]

        self._success_status_codes = success_status_codes
        self._timeout = timeout

    def is_reachable(self, url: str) -> bool:
        """
        Check if the URL is reachable.
        """
        try:
            response = requests.get(url, timeout=self._timeout)
            return response.status_code in self._success_status_codes
        except requests.RequestException:
            return False

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        urls = extract_urls(output)
        if not urls:
            return output, True, 0.0

        LOGGER.debug("Found URLs in the output", len=len(urls))

        unreachable_urls = [url for url in urls if not self.is_reachable(url)]

        if unreachable_urls:
            LOGGER.warning(f"Unreachable URLs detected", urls=unreachable_urls)
            return output, False, 1.0

        LOGGER.debug("All URLs are reachable.")
        return output, True, 0.0



"""
This plugin searches for Finicity API tokens and Client Secrets.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class FinicityDetector(RegexBasedDetector):
    """Scans for Finicity API tokens and Client Secrets."""

    secret_type = "Finicity Credentials"

    denylist = [
        # Finicity API token
        re.compile(
            r"""(?i)(?:finicity)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-f0-9]{32})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
        # Finicity Client Secret
        re.compile(
            r"""(?i)(?:finicity)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9]{20})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



"""
This plugin searches for Twitch API Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class TwitchApiTokenDetector(RegexBasedDetector):
    """Scans for Twitch API Tokens."""

    secret_type = "Twitch API Token"

    denylist = [
        re.compile(
            r"""(?i)(?:twitch)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9]{30})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        )
    ]



"""
This plugin searches for Postman API Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class PostmanApiTokenDetector(RegexBasedDetector):
    """Scans for Postman API Tokens."""

    secret_type = "Postman API Token"

    denylist = [
        re.compile(r"""(?i)\b(PMAK-[a-f0-9]{24}-[a-f0-9]{34})(?:['|\"|\n|\r|\s|\x60|;]|$)""")
    ]



from typing import List, Optional

from llm_guard.util import get_logger, lazy_load_dep

from .base import Scanner

LOGGER = get_logger(__name__)


class TokenLimit(Scanner):
    """
    A token limit scanner based on the tiktoken library. It checks if a prompt exceeds a specific token limit
    and can split a large prompt into chunks of text that each fit within the limit.
    """

    def __init__(
        self,
        *,
        limit: int = 4096,
        encoding_name: str = "cl100k_base",
        model_name: Optional[str] = None,
    ):
        """
        Initializes TokenLimit with a limit, encoding name, and model name.

        Parameters:
            limit (int): Maximum number of tokens allowed in a prompt. Default is 4096.
            encoding_name (str): Encoding model for the tiktoken library. Default is 'cl100k_base'.
            model_name (str): Specific model for the tiktoken encoding. Default is None.
        """

        self._limit = limit

        tiktoken = lazy_load_dep("tiktoken")
        if not model_name:
            self._encoding = tiktoken.get_encoding(encoding_name)
        else:
            self._encoding = tiktoken.encoding_for_model(model_name)

    def _split_text_on_tokens(self, text: str) -> (List[str], int):
        """Split incoming text and return chunks using tokenizer."""
        splits: List[str] = []
        input_ids = self._encoding.encode(text)
        start_idx = 0
        cur_idx = min(start_idx + self._limit, len(input_ids))
        chunk_ids = input_ids[start_idx:cur_idx]

        while start_idx < len(input_ids):
            splits.append(self._encoding.decode(chunk_ids))
            start_idx += self._limit
            cur_idx = min(start_idx + self._limit, len(input_ids))
            chunk_ids = input_ids[start_idx:cur_idx]

        return splits, len(input_ids)

    def scan(self, prompt: str) -> (str, bool, float):
        if prompt.strip() == "":
            return prompt, True, 0.0

        chunks, num_tokens = self._split_text_on_tokens(text=prompt)
        if num_tokens < self._limit:
            LOGGER.debug(
                "Prompt fits the maximum tokens", num_tokens=num_tokens, threshold=self._limit
            )
            return prompt, True, 0.0

        LOGGER.warning(
            "Prompt is too big. Splitting into chunks", num_tokens=num_tokens, chunks=chunks
        )

        return chunks[0], False, 1.0



"""
This plugin searches for Sentry Access Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class SentryAccessTokenDetector(RegexBasedDetector):
    """Scans for Sentry Access Tokens."""

    secret_type = "Sentry Access Token"

    denylist = [
        re.compile(
            r"""(?i)(?:sentry)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-f0-9]{64})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        )
    ]



from enum import Enum
from typing import Dict, List, Optional, Sequence, Union

from llm_guard.transformers_helpers import pipeline
from llm_guard.util import calculate_risk_score, get_logger, split_text_by_sentences

from .base import Scanner

LOGGER = get_logger(__name__)

model_path = (
    "papluca/xlm-roberta-base-language-detection",
    "ProtectAI/xlm-roberta-base-language-detection-onnx",
)


class MatchType(Enum):
    SENTENCE = "sentence"
    FULL = "full"

    def get_inputs(self, prompt: str) -> List[str]:
        if self == MatchType.SENTENCE:
            return split_text_by_sentences(prompt)

        return [prompt]


class Language(Scanner):
    """
    A Scanner subclass responsible for determining the language of a given text
    prompt and verifying its validity against a list of predefined languages.

    Note: when no languages are detected above the threshold, the prompt is considered valid.
    """

    def __init__(
        self,
        valid_languages: Sequence[str],
        *,
        threshold: float = 0.6,
        match_type: Union[MatchType, str] = MatchType.FULL,
        use_onnx: bool = False,
        transformers_kwargs: Optional[Dict] = None,
    ):
        """
        Initializes the Language scanner with a list of valid languages.

        Parameters:
            valid_languages (Sequence[str]): A list of valid language codes in ISO 639-1.
            threshold (float): Minimum confidence score.
            match_type (MatchType): Whether to match the full text or individual sentences. Default is MatchType.FULL.
            use_onnx (bool): Whether to use ONNX for inference. Default is False.
            transformers_kwargs (Optional[Dict]): Optional keyword arguments for the transformers pipeline.
        """
        if isinstance(match_type, str):
            match_type = MatchType(match_type)

        self._threshold = threshold
        self._valid_languages = valid_languages
        self._match_type = match_type

        transformers_kwargs = transformers_kwargs or {}
        transformers_kwargs["max_length"] = 512
        transformers_kwargs["truncation"] = True
        transformers_kwargs["top_k"] = None

        self._pipeline = pipeline(
            task="text-classification",
            model=model_path[0],
            onnx_model=model_path[1],
            use_onnx=use_onnx,
            **transformers_kwargs,
        )

    def scan(self, prompt: str) -> (str, bool, float):
        if prompt.strip() == "":
            return prompt, True, 0.0

        results_all = self._pipeline(self._match_type.get_inputs(prompt))
        for result_chunk in results_all:
            languages_above_threshold = [
                result["label"] for result in result_chunk if result["score"] > self._threshold
            ]

            highest_score = max([result["score"] for result in result_chunk])

            # Check if any of the languages above threshold are not valid
            if len(set(languages_above_threshold) - set(self._valid_languages)) > 0:
                LOGGER.warning(
                    "Languages are found with high confidence", languages=languages_above_threshold
                )

                return prompt, False, calculate_risk_score(highest_score, self._threshold)

        LOGGER.debug("Only valid languages are found in the text.")

        return prompt, True, 0.0



import re
from enum import Enum
from typing import List, Tuple, Union

from llm_guard.util import get_logger, lazy_load_dep
from llm_guard.vault import Vault

from .base import Scanner

LOGGER = get_logger(__name__)


class MatchingStrategy(Enum):
    """
    An enum for the different matching strategies used to find placeholders in the model output.
    """

    EXACT = "exact"
    CASE_INSENSITIVE = "case_insensitive"
    FUZZY = "fuzzy"
    COMBINED_EXACT_FUZZY = "combined_exact_fuzzy"

    @staticmethod
    def _match_exact(text: str, vault_items: List[Tuple]) -> str:
        """
        Replaces placeholders in the model output with real values from the vault.

        Parameters:
            text (str): The model output.
            vault_items (List[Tuple]): The list of items in the vault.
        """
        for vault_item in vault_items:
            LOGGER.debug("Replaced placeholder with real value", placeholder=vault_item[0])
            text = text.replace(vault_item[0], vault_item[1])

        return text

    @staticmethod
    def _match_case_insensitive(text: str, vault_items: List[Tuple]) -> str:
        """
        It replaces all the anonymized entities with the original ones
        irrespective of their letter case.

        Examples of matching:
            keanu reeves -> Keanu Reeves
            JOHN F. KENNEDY -> John F. Kennedy

        Parameters:
            text (str): The model output.
            vault_items (List[Tuple]): The list of items in the vault.
        """
        for vault_item in vault_items:
            LOGGER.debug("Replaced placeholder with real value", placeholder=vault_item[0])
            # Use regular expressions for case-insensitive matching and replacing
            text = re.sub(vault_item[0], vault_item[1], text, flags=re.IGNORECASE)

        return text

    @staticmethod
    def _match_fuzzy(text: str, vault_items: List[Tuple], max_l_dist: int = 3) -> str:
        """
        It uses fuzzy matching to find the position of the anonymized entity in the text.
        It replaces all the anonymized entities with the original ones.

        Examples of matching:
            Kaenu Reves -> Keanu Reeves
            John F. Kennedy -> John Kennedy

        Parameters:
            text (str): The model output.
            vault_items (List[Tuple]): The list of items in the vault.
        """

        fuzzysearch = lazy_load_dep("fuzzysearch")
        for vault_item in vault_items:
            LOGGER.debug("Replaced placeholder with real value", placeholder=vault_item[0])

            matches = fuzzysearch.find_near_matches(vault_item[0], text, max_l_dist=max_l_dist)

            new_text = ""
            last_end = 0
            for m in matches:
                # add the text that isn't part of a match
                new_text += text[last_end : m.start]
                # add the replacement text
                new_text += vault_item[1]
                last_end = m.end
            # add the remaining text that wasn't part of a match
            new_text += text[last_end:]
            text = new_text

        return text

    def match(self, text: str, vault_items: List[Tuple]) -> str:
        if self == MatchingStrategy.EXACT:
            return self._match_exact(text, vault_items)

        if self == MatchingStrategy.CASE_INSENSITIVE:
            return self._match_case_insensitive(text, vault_items)

        if self == MatchingStrategy.FUZZY:
            return self._match_fuzzy(text, vault_items)

        if self == MatchingStrategy.COMBINED_EXACT_FUZZY:
            text = self._match_exact(text, vault_items)
            text = self._match_fuzzy(text, vault_items)
            return text

        return text


class Deanonymize(Scanner):
    """
    A class for replacing placeholders in the model output with real values from a vault.

    This class uses the Vault class to access stored values and replaces any placeholders
    in the model's output with their corresponding values from the vault.
    """

    def __init__(
        self,
        vault: Vault,
        *,
        matching_strategy: Union[MatchingStrategy, str] = MatchingStrategy.EXACT,
    ):
        """
        Initializes an instance of the Deanonymize class.

        Parameters:
            vault (Vault): An instance of the Vault class which stores the real values.
            matching_strategy (MatchingStrategy): The strategy used to find placeholders in the model output. Defaults to the exact matching.
        """
        if isinstance(matching_strategy, str):
            matching_strategy = MatchingStrategy(matching_strategy)

        self._vault = vault
        self._matching_strategy = matching_strategy

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        vault_items = self._vault.get()
        if len(vault_items) == 0:
            LOGGER.warning("No items found in the Vault")

        output = self._matching_strategy.match(output, vault_items)

        return output, True, 0.0



"""
This plugin searches for Shippo API Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class ShippoApiTokenDetector(RegexBasedDetector):
    """Scans for Shippo API Tokens."""

    secret_type = "Shippo API Token"

    denylist = [
        re.compile(r"""(?i)\b(shippo_(live|test)_[a-f0-9]{40})(?:['|\"|\n|\r|\s|\x60|;]|$)""")
    ]



"""
This plugin searches for Hugging Face Access and Organization API Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class HuggingFaceDetector(RegexBasedDetector):
    """Scans for Hugging Face Tokens."""

    secret_type = "Hugging Face Token"

    denylist = [
        # Hugging Face Access token
        re.compile(r"""(?:^|[\\'"` >=:])(hf_[a-zA-Z]{34})(?:$|[\\'"` <])"""),
        # Hugging Face Organization API token
        re.compile(r"""(?:^|[\\'"` >=:\(,)])(api_org_[a-zA-Z]{34})(?:$|[\\'"` <\),])"""),
    ]



"""
This plugin searches for Adafruit keys
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class AdafruitKeyDetector(RegexBasedDetector):
    """Scans for Adafruit keys."""

    secret_type = "Adafruit API Key"

    denylist = [
        re.compile(
            r"""(?i)(?:adafruit)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-z0-9_-]{32})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
    ]



from .analyzer import get_analyzer, get_transformers_recognizer
from .faker import get_fake_value
from .ner_mapping import *

__all__ = [
    "get_analyzer",
    "get_transformers_recognizer",
    "get_fake_value",
    "ALL_RECOGNIZER_CONF",
    "BERT_BASE_NER_CONF",
    "BERT_LARGE_CASED_FINETUNED_COLL03_ENGLISH_CONF",
]



"""
This plugin searches for Duffel API Tokens.
"""

import re

from detect_secrets.plugins.base import RegexBasedDetector


class DuffelApiTokenDetector(RegexBasedDetector):
    """Scans for Duffel API Tokens."""

    secret_type = "Duffel API Token"

    denylist = [
        # Duffel API Token
        re.compile(r"""(?i)duffel_(test|live)_[a-z0-9_\-=]{43}"""),
    ]



"""
This plugin searches for Databricks API token.
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class DatabricksApiTokenDetector(RegexBasedDetector):
    """Scans for Databricks API token."""

    secret_type = "Databricks API Token"

    denylist = [
        re.compile(r"""(?i)\b(dapi[a-h0-9]{32})(?:['|\"|\n|\r|\s|\x60|;]|$)"""),
    ]



"""
This plugin searches for Clojars API tokens
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class ClojarsApiTokenDetector(RegexBasedDetector):
    """Scans for Clojars API tokens."""

    secret_type = "Clojars API token"

    denylist = [
        # For Clojars API token
        re.compile(r"(?i)(CLOJARS_)[a-z0-9]{60}"),
    ]



from typing import Tuple

from llm_guard.exception import LLMGuardValidationError
from llm_guard.transformers_helpers import get_tokenizer, is_onnx_supported
from llm_guard.util import device, get_logger, lazy_load_dep

from .base import Scanner

LOGGER = get_logger(__name__)

MODEL_EN_BGE_BASE = (
    "BAAI/bge-base-en-v1.5",
    "zeroshot/bge-base-en-v1.5-quant",  # Quantized and converted to ONNX version of BGE base
)
MODEL_EN_BGE_LARGE = (
    "BAAI/bge-large-en-v1.5",
    "zeroshot/bge-large-en-v1.5-quant",  # Quantized and converted to ONNX version of BGE large
)
MODEL_EN_BGE_SMALL = (
    "BAAI/bge-small-en-v1.5",
    "zeroshot/bge-small-en-v1.5-quant",  # Quantized and converted to ONNX version of BGE small
)

all_models = [MODEL_EN_BGE_LARGE, MODEL_EN_BGE_BASE, MODEL_EN_BGE_SMALL]

torch = lazy_load_dep("torch")
np = lazy_load_dep("numpy")


class Relevance(Scanner):
    """
    A class used to scan the relevance of the output of a language model to the input prompt.

    This class uses SentenceTransformers to encode the prompt and output into vector embeddings, then computes
    the cosine similarity between them. If the similarity is below a given threshold, the output is considered
    not relevant to the prompt.
    """

    def __init__(
        self, *, threshold: float = 0.5, model: Tuple = MODEL_EN_BGE_BASE, use_onnx: bool = False
    ):
        """
        Initializes an instance of the Relevance class.

        Parameters:
            threshold (float): The minimum similarity score to compare prompt and output.
            model (Tuple): Model for calculating embeddings. Default is `BAAI/bge-base-en-v1.5`.
            use_onnx (bool): Whether to use the ONNX version of the model. Defaults to False.
        """

        self._threshold = threshold

        if model not in all_models:
            raise LLMGuardValidationError(f"Model must be in the list of allowed: {all_models}")

        model_path = model[0]

        self.pooling_method = "cls"
        self.normalize_embeddings = True

        if use_onnx and is_onnx_supported() is False:
            LOGGER.warning("ONNX is not supported on this machine. Using PyTorch instead of ONNX.")
            use_onnx = False

        if use_onnx:
            model_path = model[1]
            optimum_onnxruntime = lazy_load_dep(
                "optimum.onnxruntime",
                "optimum[onnxruntime-gpu]" if device().type == "cuda" else "optimum[onnxruntime]",
            )
            self._model = optimum_onnxruntime.ORTModelForFeatureExtraction.from_pretrained(
                model_path,
                export=False,
                provider="CUDAExecutionProvider"
                if device().type == "cuda"
                else "CPUExecutionProvider",
                use_io_binding=True if device().type == "cuda" else False,
            )
            LOGGER.debug("Initialized ONNX model", model=model_path, device=device())
        else:
            transformers = lazy_load_dep("transformers")
            self._model = transformers.AutoModel.from_pretrained(model_path).to(device())
            LOGGER.debug("Initialized model", model=model_path, device=device())
            self._model.eval()

        self._tokenizer = get_tokenizer(model_path)

    def pooling(self, last_hidden_state: torch.Tensor, attention_mask: torch.Tensor = None):
        if self.pooling_method == "cls":
            return last_hidden_state[:, 0]
        elif self.pooling_method == "mean":
            s = torch.sum(last_hidden_state * attention_mask.unsqueeze(-1).float(), dim=1)
            d = attention_mask.sum(dim=1, keepdim=True).float()
            return s / d

    @torch.no_grad()
    def _encode(self, sentence: str, max_length: int = 512) -> np.ndarray:
        inputs = self._tokenizer(
            [sentence],
            padding=True,
            truncation=True,
            return_tensors="pt",
            max_length=max_length,
        ).to(device())
        last_hidden_state = self._model(**inputs, return_dict=True).last_hidden_state
        embeddings = self.pooling(last_hidden_state, inputs["attention_mask"])
        if self.normalize_embeddings:
            embeddings = torch.nn.functional.normalize(embeddings, dim=-1)
        embeddings = embeddings.cpu().numpy()

        return embeddings[0]

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        if output.strip() == "":
            return output, True, 0.0

        prompt_embedding = self._encode(prompt)
        output_embedding = self._encode(output)
        similarity = prompt_embedding @ output_embedding.T

        if similarity < self._threshold:
            LOGGER.warning("Result is not similar to the prompt", similarity_score=similarity)

            return output, False, round(1 - similarity, 2)

        LOGGER.debug("Result is similar to the prompt", similarity_score=similarity)

        return output, True, 0.0



from llm_guard.input_scanners.ban_topics import MODEL_BASE
from llm_guard.transformers_helpers import get_tokenizer_and_model_for_classification
from llm_guard.util import device, get_logger, lazy_load_dep

from .base import Scanner

torch = lazy_load_dep("torch")
LOGGER = get_logger(__name__)

_model = MODEL_BASE


class FactualConsistency(Scanner):
    """
    FactualConsistency Class:

    This class checks for entailment between a given prompt and output using a pretrained NLI model.
    """

    def __init__(self, *, minimum_score=0.5, use_onnx=False):
        """
        Initializes an instance of the Refutation class.

        Parameters:
            minimum_score (float): The minimum entailment score for the output to be considered valid. Defaults to 0.5.
            use_onnx (bool): Whether to use the ONNX version of the model. Defaults to False.
        """

        self._minimum_score = minimum_score

        self._tokenizer, self._model = get_tokenizer_and_model_for_classification(
            model=_model["path"],
            onnx_model=_model["onnx_path"],
            use_onnx=use_onnx,
        )
        self._model = self._model.to(device())

    def scan(self, prompt: str, output: str) -> (str, bool, float):
        if prompt.strip() == "":
            return output, True, 0.0

        tokenized_input_seq_pair = self._tokenizer(
            output, prompt, padding=True, truncation=True, return_tensors="pt"
        ).to(device())

        model_output = self._model(
            tokenized_input_seq_pair["input_ids"], tokenized_input_seq_pair["attention_mask"]
        )
        model_prediction = torch.softmax(model_output["logits"][0], -1).tolist()
        label_names = ["entailment", "not_entailment"]
        prediction = {
            name: round(float(pred), 2) for pred, name in zip(model_prediction, label_names)
        }

        entailment_score = prediction["entailment"]
        if entailment_score < self._minimum_score:
            LOGGER.warning("Entailment score is below the threshold", prediction=prediction)

            return output, False, prediction["not_entailment"]

        LOGGER.debug("The output is factually consistent", prediction=prediction)

        return output, True, 0.0



"""
This plugin searches for Adobe keys
"""
import re

from detect_secrets.plugins.base import RegexBasedDetector


class AdobeSecretDetector(RegexBasedDetector):
    """Scans for Adobe client keys."""

    secret_type = "Adobe Client Keys"

    denylist = [
        # Adobe Client ID (OAuth Web)
        re.compile(
            r"""(?i)(?:adobe)(?:[0-9a-z\-_\t .]{0,20})(?:[\s|']|[\s|"]){0,3}(?:=|>|:{1,3}=|\|\|:|<=|=>|:|\?=)(?:'|\"|\s|=|\x60){0,5}([a-f0-9]{32})(?:['|\"|\n|\r|\s|\x60|;]|$)"""
        ),
        # Adobe Client Secret
        re.compile(r"(?i)\b((p8e-)[a-z0-9]{32})(?:['|\"|\n|\r|\s|\x60|;]|$)"),
    ]
